/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.BufferedOutputStream;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.util.Hashtable;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.Properties;
import java.util.Enumeration;
import java.util.Locale;
import java.net.URL;
import java.net.MalformedURLException;
import java.lang.reflect.InvocationTargetException;

public class Main extends ClassLoader {

   private static final String CUBE_JAR_FILE_BASE = "cube";
   private static final String CUBE_JAR_FILE_EXTENSION = ".jar";
   private static final String FLASH_URL_FILE = "flash.url";

   private static final String SYSTEM_PROPERTY_PROPERTIES_FILE = "com.opensugar.cube.propertiesFile";

   public static void main( String[] args ) throws Exception {
      new Main( args );
   }

   private int n;
   private File libDir;
   private File jarFile;
   private File flashURLFile;

   private ZipFile zipFile;
   private Hashtable classCache;

   protected Main() {
   }

   protected Main( String[] args ) throws Exception {
      Object splashScreen = null;
      if ( !System.getProperty( "com.opensugar.cube.splashScreen", "" + true ).equals( "" + false ) ) {
         try {
            splashScreen = new SplashScreen();
         }
         catch ( Throwable ignore ) {}
      }

      IOException exceptionWhileLoadingProperties = null;
      try {
         loadProperties();
      }
      catch( IOException e ) {
         exceptionWhileLoadingProperties = e;
      }

      FileNotFoundException exceptionWhileSettingSystemStreams = null;
      try {
         setSystemStreams();
      }
      catch ( FileNotFoundException e ) {
         exceptionWhileSettingSystemStreams = e;
      }

      // Now we can start writing to stderr and stdout, NOT before the call to
      // setSystemStreams has returned

      setLanguage();

      if ( exceptionWhileLoadingProperties != null ) {
         System.err.println( "Error while loading properties: " + exceptionWhileLoadingProperties.toString() );
      }
      if ( exceptionWhileSettingSystemStreams != null ) {
         System.err.println( "Error while setting system streams: " + exceptionWhileSettingSystemStreams.toString() );
      }

      libDir = new File( System.getProperty( "com.opensugar.cube.libDir", "lib" ) );
      flashURLFile = new File( "flash.url" );
      boolean done = false;
      File oldJarFile = null;
      while ( !done ) {
         n = findCubeJarFileIndex( libDir );
         jarFile = new File( libDir, CUBE_JAR_FILE_BASE + n + CUBE_JAR_FILE_EXTENSION );
         System.getProperties().put( "com.opensugar.cube.systemBundleJarFile", jarFile.getAbsolutePath() );

         // Modification for Sony
         // They start the cube from some strange sort of process which calls the main
         // method with a null argument
         if ( args == null ) {
            args = new String[ 0 ];
         }

         Class[] argTypes = { args.getClass(), Object.class };
         Object[] args2 = { args, splashScreen };
         try {
            classCache = new Hashtable();
            zipFile = new ZipFile( jarFile );
            Object returnValue = loadClass( "com.opensugar.cube.Launcher" ).getMethod( "main", argTypes ).invoke( null, args2 );
            if ( System.getProperty( "com.opensugar.cube.classloader.type", "dynamic" ).equals( "basic" ) ) {
               return;
            }
// problem: old jar file does not get deleted until cube exits
            if ( oldJarFile != null && !oldJarFile.delete() ) {
               System.err.println( "Warning: Could not delete old cube jar file: " + oldJarFile.getAbsolutePath() );
            }
         }
         catch ( Throwable e ) {
            if ( e instanceof InvocationTargetException ) {
               e = ( (InvocationTargetException)e ).getTargetException();
            }
            if ( oldJarFile != null ) {
               // error while starting cube after a flash operation
               // restore old cube version
               e.printStackTrace();
               System.out.println( "ERROR: could not start updated cube version" );
               System.out.println( "       will try to restart with old version" );
               jarFile.delete();
               jarFile = oldJarFile;
               oldJarFile = null;
               continue;
            }
            else {
               e.printStackTrace();
            }
         }

         if ( flashURLFile.exists() ) {
            try {
               oldJarFile = flashCube();
               System.out.println( "Cube flashed" );
            }
            catch ( Exception e ) {
               System.out.println( "Could not flash cube" );
               e.printStackTrace();
            }
            finally {
               if ( !flashURLFile.delete() ) {
                  System.out.println( "ERROR: Could not delete flash URL file: " + flashURLFile.getAbsolutePath() );
                  System.exit( 1 );
               }
            }
         }
         else {
            done  = true;
         }
      }

      System.exit( 0 );
   }

   // Load properties from the properties file: cube.properties
   protected void loadProperties() throws IOException {
      String propertiesFileName = System.getProperty( SYSTEM_PROPERTY_PROPERTIES_FILE, "cube.properties" );
      File propertiesFile = new File( propertiesFileName );
      if ( propertiesFile.exists() && propertiesFile.isFile() ) {
         Properties properties = new Properties();
         properties.load( new FileInputStream( propertiesFile ) );
         Properties systemProperties = System.getProperties();
         Enumeration keys = properties.keys();
         Object key;
         while ( keys.hasMoreElements() ) {
            key = keys.nextElement().toString().trim();
            systemProperties.put( key, properties.get( key ).toString().trim() );
         }
      }
   }

   // if com.opensugar.cube.err property is null, then leave standard error stream alone
   // if                                 = "null", then mute standard error stream
   // if                                 = "file:xxxxx", then redirect standard error stream to a file named xxxxx
   // same for output stream
   // if com.opensugar.cube.err=com.opensugar.cube.out, then use the same stream for
   // both standard error and output stream (do not create the same stream twice)
   protected void setSystemStreams() throws FileNotFoundException {
      String filePrefix = "file:";
      String stderrString = System.getProperty( "com.opensugar.cube.stderr" );
      OutputStream stderr = null;
      if ( stderrString != null ) {
         if ( stderrString.equals( "null" ) ) {
            stderr = new NullOutputStream();
         }
         if ( stderrString.startsWith( filePrefix ) && stderrString.length() > ( filePrefix ).length() ) {
            stderr = new FileOutputStream( stderrString.substring( filePrefix.length() ) );
         }
      }
      String stdoutString = System.getProperty( "com.opensugar.cube.stdout" );
      OutputStream stdout = null;
      if ( stdoutString == null ) {
         if ( stderrString == null ) {
            stdout = stderr;
         }
      }
      else {
         if ( stdoutString.equals( stderrString ) ) {
            stdout = stderr;
         }
         else if ( stdoutString.equals( "null" ) ) {
            stdout = new NullOutputStream();
         }
         else if ( stdoutString.startsWith( filePrefix ) && stdoutString.length() > ( filePrefix ).length() ) {
            stdout = new FileOutputStream( stdoutString.substring( filePrefix.length() ) );
         }
      }
      if ( stderr != null ) {
         System.setErr( new PrintStream( stderr ) );
      }
      if ( stdout != null ) {
         System.setOut( new PrintStream( stdout ) );
      }

      String stdinString = System.getProperty( "com.opensugar.cube.stdin" );
      InputStream stdin = null;
      if ( stdinString != null ) {
         if ( stdinString.equals( "null" ) ) {
            stdin = new NullInputStream();
         }
         else if ( stdinString.startsWith( filePrefix ) && stdinString.length() > ( filePrefix ).length() ) {
            File file = new File( stdinString.substring( filePrefix.length() ) );
            if ( file.exists() ) {
               stdin = new FileInputStream( file );
            }
            else {
               stdin = new NullInputStream();
            }
         }
      }
      if ( stdin != null ) {
         System.setIn( stdin );
      }
   }

   private void setLanguage() {
      String language = System.getProperty( "com.opensugar.cube.language", "" ).toLowerCase();
      if ( language.equals( "fr" ) ) {
         Locale.setDefault( Locale.FRANCE );
      }
      else if ( language.equals( "uk" ) ) {
         Locale.setDefault( Locale.UK );
      }
      else if ( language.equals( "de" ) ) {
         Locale.setDefault( Locale.GERMANY );
      }
      else if ( language.equals( "it" ) ) {
         Locale.setDefault( Locale.ITALY );
      }
      else {
         Locale.setDefault( Locale.US );
      }
   }

   // download new cube jar file
   // return old cube jar file so it can be deleted if restart successful
   private File flashCube() throws MalformedURLException, IOException {
      File oldJarFile = jarFile;
      String location = readFile( flashURLFile ).trim();
      URL url;
      if ( location.indexOf( "://" ) == -1 ) {
         url = new URL( "file://" + location );
      }
      else {
         url = new URL( location );
      }
      n++;
      jarFile = new File( libDir, CUBE_JAR_FILE_BASE + n + CUBE_JAR_FILE_EXTENSION );
      transferData( url.openStream(), new FileOutputStream( jarFile ) );
      return oldJarFile;
   }

   private int findCubeJarFileIndex( File libDir ) {
      if ( !libDir.exists() ) {
         System.out.println( "FATAL ERROR: cube lib dir does not exist: " + libDir.getAbsolutePath() );
         System.exit( 1 );
      }

      String[] listing = libDir.list();
      int largestN = -1;
      for ( int i = 0; i < listing.length; i++ ) {
         if ( listing[ i ].startsWith( CUBE_JAR_FILE_BASE ) && listing[ i ].endsWith( CUBE_JAR_FILE_EXTENSION ) ) {
            String nString = listing[ i ].substring( CUBE_JAR_FILE_BASE.length(), listing[ i ].lastIndexOf( CUBE_JAR_FILE_EXTENSION ) );
            try {
               int n = Integer.valueOf( nString ).intValue();
               if ( n > largestN ) {
                  largestN = n;
               }
            }
            catch ( NumberFormatException e ) {}
         }
      }
      if ( largestN == -1 ) {
         System.out.println( "FATAL ERROR: cannot find cube jar file" );
         System.exit( 1 );
      }

      return largestN;
   }

   private String readFile( File file ) throws IOException {
      FileInputStream fis = null;
      try {
         fis = new FileInputStream( file );
         int n = fis.available();
         byte[] data = new byte[ n ];
         fis.read( data );
         return new String( data );
      }
      catch( IOException e ) {
         throw e;
      }
      finally {
         try {
            fis.close();
         }
         catch( Exception ignore ) {}
      }
   }

   private void transferData( InputStream source, OutputStream destination ) throws IOException {
      BufferedInputStream bis = new BufferedInputStream( source );
      BufferedOutputStream bos = new BufferedOutputStream( destination );
      byte[] buffer = new byte[ 4096 ];
      int n;
      IOException exception = null;
      try {
         while ( ( n = bis.read( buffer ) ) != -1 ) {
            bos.write( buffer, 0, n );
         }
      }
      catch( IOException e ) {
         exception = e;
      }
      finally {
         bis.close();
         bos.close();
      }

      if ( exception != null ) {
         throw exception;
      }
   }

   public synchronized Class loadClass( String name, boolean resolve ) throws ClassNotFoundException {
      try {
         return Class.forName( name );
      }
      catch ( ClassNotFoundException ignore ) {}

      Class c = (Class)classCache.get( name );
      if (c == null) {
         try {
            byte data[] = getJarEntryContents( name.replace( '.', '/' ) + ".class" );
            if ( data == null ) {
               return super.loadClass( name, resolve );
            }
            else {
               c = defineClass( data, 0, data.length );
               classCache.put( name, c );
            }
         }
         catch ( IOException e ) {
System.out.println( e );
            throw new RuntimeException( "IOException while loading class: " + name + "\n" + e.toString() );
         }
      }
      if ( resolve ) {
         resolveClass( c );
      }
      return c;
   }

   private byte[] getJarEntryContents( String entryName ) throws IOException {
      ZipEntry entry = zipFile.getEntry( entryName );
      if ( entry != null ) {
         int entrySize = (int)entry.getSize();
         if ( entrySize < 0 ) {
            throw new IOException( "Unknown size for " + entry );
         }
         InputStream is = zipFile.getInputStream( entry );
         byte[] bytes = new byte[ entrySize ];
         DataInputStream dis = new DataInputStream( is );
         dis.readFully( bytes );
         is.close();
         return bytes;
      }
      return null;
   }

   private class NullOutputStream extends OutputStream {
      public void close() {}
      public void flush() {}
      public void write( byte[] b ) {}
      public void write( byte[] b, int off, int len ) {}
      public void write( int b ) {}
   }

   private class NullInputStream extends InputStream {
      public int available() {
         return 0;
      }
      public void close() {}
      public void mark( int readlimit ) {}
      public boolean markSupported() {
         return false;
      }
      public int read() {
         return -1;
      }
      public int read( byte[] b ) {
         return -1;
      }
      public int read( byte[] b, int off, int len ) {
         return -1;
      }
      public void reset() {}
      public long skip( long n ) {
         return 0;
      }
   }

   private class SplashScreen extends java.awt.Window implements java.util.Observer {
      private SplashImageCanvas imageCanvas;
      private SplashProgressCanvas progressCanvas;
      public SplashScreen() {
         super( new java.awt.Frame() );
         int progressHeight = 4;
         int imageHeight = 240;
         int width = 320;
         imageCanvas = new SplashImageCanvas( System.getProperty( "com.opensugar.cube.splashMessage", "Cube starting..." ) );
         progressCanvas = new SplashProgressCanvas( width, progressHeight );
         setLayout( new java.awt.BorderLayout() ) ;
         add( imageCanvas,java.awt.BorderLayout.CENTER );
         add( progressCanvas, java.awt.BorderLayout.SOUTH );
         java.awt.Dimension size = new java.awt.Dimension( width, imageHeight + progressHeight );
         java.awt.Toolkit toolkit = java.awt.Toolkit.getDefaultToolkit();
         java.awt.Dimension screenSize = toolkit.getScreenSize();
         setSize( size );
         setLocation( ( screenSize.width - size.width ) / 2, ( screenSize.height - size.height ) / 2 );
         show();
         java.awt.Image image = toolkit.getImage( System.getProperty( "com.opensugar.cube.splashImage", "cube.jpg" ) );
         java.awt.MediaTracker tracker = new java.awt.MediaTracker( this );
         tracker.addImage( image, 0 );
         try {
            tracker.waitForID( 0, 5000 );
         }
         catch ( InterruptedException ignore ) {}
         imageCanvas.setImage( image );
      }
      public void update( java.util.Observable observable, Object arg ) {
         progressCanvas.setProgress( ( (Integer)arg ).intValue() );
      }
   }

   private class SplashImageCanvas extends java.awt.Canvas {
      private String message;
      private java.awt.Image image;
      public SplashImageCanvas( String message ) {
         this.message = message;
      }
      public void setImage( java.awt.Image image ) {
         if ( image.getWidth( this ) > 0 ) {
            this.image = image;
            repaint( 50 );
            try {
               Thread.sleep( 100 );
            }
            catch ( InterruptedException ignore ) {}
         }
      }
      public void paint( java.awt.Graphics g ) {
         java.awt.Dimension size = getSize();
         g.setColor( java.awt.Color.black );
         if ( image == null ) {
            g.fillRect( 0, 0, size.width - 1, size.height - 1 );
            g.setColor( java.awt.Color.lightGray );
            g.drawString( message, ( size.width - g.getFontMetrics().stringWidth( message ) ) / 2, size.height / 2 );
         }
         else {
            g.drawImage( image, 0, 0, this );
            g.drawRect( 0, 0, size.width - 1, size.height - 1 );
            g.drawString( message, ( size.width - g.getFontMetrics().stringWidth( message ) ) / 2, size.height - 2 );
         }
      }
   }

   private class SplashProgressCanvas extends java.awt.Canvas {
      private int progress;
      private java.awt.Dimension size;
      private boolean firstPaint = true;
      public SplashProgressCanvas( int width, int height ) {
         size = new java.awt.Dimension( width, height );
         setBackground( java.awt.Color.lightGray );
      }
      public java.awt.Dimension preferredSize() {
         return size;
      }
      public void setProgress( int progress ) {
         this.progress = progress;
         repaint( 50 );
         try {
            Thread.sleep( 100 );
         }
         catch ( InterruptedException ignore ) {}
      }
      public void paint( java.awt.Graphics g ) {
//         if ( firstPaint ) {
//            g.setColor( java.awt.Color.lightGray );
//            g.fillRect( 0, 0, size.width, size.height );
//            firstPaint = false;
//         }
         g.setColor( java.awt.Color.blue );
         int prog = ( size.width - 1 ) * progress / 100;
         g.fillRect( 0, 0, prog, size.height );
      }
   }

}
