/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube;

import com.opensugar.cube.java2.PackageHelper;
import com.opensugar.cube.packageAdmin.PackageAdminImpl;
import com.opensugar.cube.packageAdmin.PackageImport;
import com.opensugar.cube.serviceRegistry.ServiceRegistry;
import com.opensugar.cube.serviceRegistry.ServiceRegistryException;
import com.opensugar.cube.ldap.LDAPFilter;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.FrameworkListener;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.BundleListener;
import org.osgi.framework.SynchronousBundleListener;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.Constants;
import org.osgi.framework.ServicePermission;
import org.osgi.service.permissionadmin.PermissionAdmin;

import java.io.File;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Locale;
import java.util.Observer;

// Abstract engine.
// Cannot be instantiated.
// Should be sub-classed to define engines that are appropriate for different
// operating conditions.
public abstract class AbstractCube {

   public static final String VERSION = "2.1.3";
   public static final String VENDOR = "Opensugar S.A.";
   public static final String OSGI_RELEASE = "2";

   private static final String SYSTEM_PROPERTY_CUBE_VERSION = "com.opensugar.cube.version";
   private static final String SYSTEM_PROPERTY_CUBE_DIR = "com.opensugar.cube.dir";
//   private static final String SYSTEM_PROPERTY_PERMANENT_STORAGE_DIR = "com.opensugar.cube.permanentStorageDir";
   private static final String SYSTEM_PROPERTY_SYSTEM_BUNDLE_JAR_FILE = "com.opensugar.cube.systemBundleJarFile";
   private static final String SYSTEM_PROPERTY_CUBE_STARTED_FLAG_FILE = "com.opensugar.cube.startedFlagFile";

   private static final String FLASH_URL_FILE = "flash.url";

   private static final String defaultCubeDir = "osc/bs";
   private static final String defaultPermanentStorageDir = "osc/.opensugar";
   private static final String defaultSystemBundleJarFile = "systemBundle.jar";


   protected static final String bundleDirNameBase = "b";
   protected static final String bundleJarNameBase = "bundle";
   protected static final String bundleJarNameExtension = ".jar";
   protected static final String bundleLocationFileName = "location";
   protected static final String bundleUninstalledFlagFileName = "uninstalled.flag";
   protected static final String bundleStartedFlagFileName = "started.flag";
   protected static final String bundleInnerFilesDirName = "inner";
   protected static final String bundleDataDirName = "data";
   protected static final String bundleIdFileName = ".bid";

   // If an attempt is made to start or stop a bundle that is in the stating or
   // stopping state, the bundle needs to wait until the bundle state changes.
   // If this does not happen in a reasonable time however, the bundle must give up.
   private int reasonableTimeToWait = 5000;

   // list of bundles in this engine   // bundle id --> bundle
   private Hashtable id_bundle;
   // package admin service for this cube
   private PackageAdminImpl packageAdmin;
   // service registry for this engine
   private ServiceRegistry serviceRegistry;

   // list of service event listeners
   private ServiceListeners serviceListeners;
   // list of synchrnous bundle listeners;
   private BundleListeners bundleListeners;
   // list of framework event listeners
   private FrameworkListeners frameworkListeners;
   // flag to determine whether or not framework is firing events
   private boolean handlingEvents = false;

   // the largest bundle id of any bundle that has been installed on the current run of the cube
   private long largestBundleId = -1;

   // hashtable of directories used by bundles to store inner files contained in the bundle jar file
   private Hashtable bundleInnerFileDirectories;

   // framework state
   //    can be   * Bundle.RESOLVED before framework is started
   //             * Bundle.STARTING while the framework is staring
   //             * Bundle.ACTIVE while the framework is running
   //             * Bundle.STOPPING while the framework is shutting down
   private int state = Bundle.RESOLVED;

   // framework working directory
   private File baseDirectory;

// *****************************************************************************
// Constructor, startup, shutdown, version
// *****************************************************************************

   protected AbstractCube() {
      this( null );
   }

   protected AbstractCube( File baseDirectory ) {
      // Here, we do things that need to be done once only, at engine initialization
      this.baseDirectory = baseDirectory;

      System.out.println( "Opensugar cube " + VERSION );
      System.out.println( "Implements OSGi Service Platform, Specification Release " + OSGI_RELEASE );

      System.out.println( "Copyright (c) 2001, Opensugar S.A. All Rights Reserved." );
      System.out.println( "Opensugar is a Registered Trademark." );
      System.out.println( "" );

      // Set cube vendor and version property
      System.getProperties().put( Constants.FRAMEWORK_VENDOR, getProperty( Constants.FRAMEWORK_VENDOR ) );
      System.getProperties().put( Constants.FRAMEWORK_VERSION, getProperty( Constants.FRAMEWORK_VERSION ) );

      // Set cube version property
      System.getProperties().put( SYSTEM_PROPERTY_CUBE_VERSION, VERSION );

      // Set permanent storage dir property as it may be needed by bundles that want
      // to store permanent data
//      if ( System.getProperty( SYSTEM_PROPERTY_PERMANENT_STORAGE_DIR ) == null ) {
//         System.getProperties().put( SYSTEM_PROPERTY_PERMANENT_STORAGE_DIR, defaultPermanentStorageDir );
//      }

      initLog();
   }

   public void flash( String flashURL ) throws IOException {
      writeFile( new File( FLASH_URL_FILE ), flashURL );
      shutdown( false );
   }

   public synchronized void startup( Observer progressListener ) {
      // Here, we do things that need to be done on every engine startup

      // clean-up from previous runs
      deleteUninstalledBundleDirectories();
      deleteObsoleteBundleFiles();

      // Enable event handling
      serviceListeners = new ServiceListeners();
      bundleListeners = new BundleListeners();
      frameworkListeners = new FrameworkListeners();
      handlingEvents = true;

      // enter starting state
      state = Bundle.STARTING;

      // Initialize hash table of bundle contexts
      id_bundle = new Hashtable();

      // Initialize hash table of bundle inner jar directories
      bundleInnerFileDirectories = new Hashtable();

      // Initialize package admin
      packageAdmin = new PackageAdminImpl( this );

      // Initialize service registry
      serviceRegistry = new ServiceRegistry( this );

      // Create engine base directory if it does not exists.
      getBaseDirectory().mkdirs();
      // Install system bundle
      installSystemBundle( new File( System.getProperty( SYSTEM_PROPERTY_SYSTEM_BUNDLE_JAR_FILE, defaultSystemBundleJarFile ) ) );

      // Reinstall bundles that were installed on last engine run
      //
      // OSGi 2 spec does not specify how to handle errors that may occur while
      // reinstalling bundles
      reinstallBundles();
      // Refresh packages from all bundles (except system bundle) so that, no matter
      // what order the bundles are installed in, we use the latest possible package
      // versions, when the cube is restarted
      Bundle[] bs = getBundles();
      Vector v = new Vector();
      for ( int i = 0; i < bs.length; i++ ) {
         v.addElement( bs[ i ] );
      }
      v.removeElement( getBundle( (long)0 ) );
      bs = new Bundle[ v.size() ];
      v.copyInto( bs );
      // refresh package synchronously
      packageAdmin.doRefreshPackages( bs );
      // Restart bundles that were active on last engine run
      //
      // OSGi 2 spec says that BundleException while restarting bundles should be
      // broadcast as framework error events
      restartBundles( progressListener );

      // enter active state
      state = Bundle.ACTIVE;

      fireFrameworkStartedEvent();

      System.out.println( "Cube started" );

      String startedFlagFile = System.getProperty( SYSTEM_PROPERTY_CUBE_STARTED_FLAG_FILE );
      if ( startedFlagFile != null ) {
         File file = new File( startedFlagFile );
         try {
            writeFile( file, "" );
         }
         catch ( IOException e ) {
            log( LOG_ERROR, "Unable to create started flag file: " + file.getAbsolutePath(), e );
         }
      }
   }

   public synchronized void shutdown( boolean exitJVM ) {
      // enter stopping state
      state = Bundle.STOPPING;

      // suspend active bundles
      //
      // OSGi 2 spec says that BundleException while suspending bundles should be
      // broadcast as framework error events
      BundleImpl[] bundles = getBundles();
      for ( int i = bundles.length - 1; i >= 0; i-- ) {
         if ( bundles[ i ].getBundleId() != 0 && bundles[ i ].getState() == bundles[ i ].ACTIVE ) {
            try {
               bundles[ i ].stop( true );
            }
            catch( BundleException e ) {
               fireFrameworkEvent( FrameworkEvent.ERROR, bundles[ i ], e );
            }
         }
      }

      // Disable event handling
      handlingEvents = false;

      state = Bundle.RESOLVED;

      System.out.println( "Cube stopped" );

      // Note: no data storage or updates to bundle directories are made in this method
      // This is good because it means that if the engine is stopped brutally or crashed,
      // the fact that shutdown() has not been called will not be a problem when the engine
      // is restarted.

      if ( exitJVM ) {
         System.exit( 0 );
      }
   }

   protected abstract boolean enforcesPermissions();

// *****************************************************************************
// Bundle operations
// *****************************************************************************

   protected Bundle reinstallBundle( long id ) throws BundleException, IOException {
      File jarFile = getCurrentBundleJarFile( id );
      String location = readStoredBundleLocation( id );
      return installBundle( id, location, jarFile );
   }

   public Bundle installBundle( String location ) throws BundleException {
      Util.checkNullParameter( location, getClass().getName(), "installBundle", "location" );
      return installBundle( location, null );
   }

   protected Bundle installBundle( String location, InputStream in ) throws BundleException {
      Util.checkNullParameter( location, getClass().getName(), "installBundle", "location" );

      // If a bundle with the same location string has already been installed
      // into the framework, return that bundle
      try {
         Bundle bundle = getBundle( location );
         if ( bundle != null && bundle.getState() != bundle.UNINSTALLED ) {
            return getBundle( location );
         }
      }
      catch( IllegalArgumentException ignore ) {
         // bundle with this location has not already been installed
      }

      // Generate id for the bundle
      long id = generateNewBundleId();

      // With the new bundle id generated, we should be able to create a fresh directory.
      // However there can be a problem (observed with windows NT)
      // Start an engine, open a file explorer and look at the contents of the bundle with the
      // largest id, b3 say.
      // Then uninstall bundle 3, stop the engine and restart it.
      // The directory for bundle b3 is deleted when the engine is restarted (since the bundle
      // has been uninstalled). However the file explorer still has some sort of hook on that
      // directory.
      // When we create directory b3 now, it seems to be ok but when we call list() on the
      // directory, we get null, and this causes exceptions.
      // To avoid this problem, we test if this happens, and if it does, we just take another
      // id.
      File bundleDirectory = null;
      while ( bundleDirectory == null ) {
         // Create directory for bundle
         bundleDirectory = createFreshBundleDirectory( id );
         if ( bundleDirectory.list() == null ) {
            bundleDirectory = null;
            id++;
         }
      }

      try {
         // Retreive jar file and store in bundle directory
         if ( in == null ) {
            URL url = new URL( location );
            in = url.openStream();
         }
         File bundleJar = getNewBundleJarFile( id );
         Util.transferData( in, new FileOutputStream( bundleJar ) );

         // Store bundle location in permanent storage
         if ( id != 0 ) {
            recordBundleLocation( id, location );
         }

         return installBundle( id, location, bundleJar );
      }
      catch( Exception e ) {
         // Undo any work performed, and rethrow exception
         if ( !Util.recursiveFileDelete( bundleDirectory ) ) {
            log( LOG_WARNING, "Unable to delete bundle directory after bundle installation failure" );
         }

         if ( e instanceof BundleException ) {
            throw (BundleException)e;
         }
         else {
            throw new BundleException( "Unable to install bundle: " + location, e );
         }
      }
   }

   private Bundle installSystemBundle( File bundleJar ) {
      try {
         if ( !bundleJar.exists() ) {
            throw new BundleException( "System bundle jar file not found: " + bundleJar.getAbsolutePath() );
         }
         return installBundle( (long)0, Constants.SYSTEM_BUNDLE_LOCATION, bundleJar );
      }
      catch( BundleException e ) {
//         fireFrameworkErrorEvent( null, e );
         Throwable rootCause = e;
         if ( e.getNestedException() != null ) {
            rootCause = e.getNestedException();
         }
         log( LOG_ERROR, "FATAL ERROR: System bundle could not be installed", rootCause );

         System.exit( 1 );
         return null;
      }
   }

   // This method is called directly by the framework to install local copy of bundle
   // on framework restarts.
   // It is also called by this class's installBundle( location, in ) to complete a new
   // bundle installation.
   private Bundle installBundle( long id, String location, File bundleJar ) throws BundleException {
      if ( id > largestBundleId ) {
         largestBundleId = id;
      }

      try {
         // Create bundle
         // Allocate associated resources
         // Set state to INSTALLED
         // Resolve classpath and native code dependencies if any
         BundleImpl bundle;
         if ( id == 0 ) {
            bundle = new SystemBundle( this, bundleJar );
         }
         else {
            bundle = new BundleImpl( this, id, location, bundleJar );
         }

         // Broadcast bundle installed event
         fireBundleEvent( BundleEvent.INSTALLED, bundle );

         id_bundle.put( new Long( bundle.getBundleId() ), bundle );
         packageAdmin.resolveBundles();

         log( LOG_INFO, bundle.getLocation() + " installed (bundle id: " + bundle.getBundleId() + ")" );

         // Return bundle
         return bundle;
      }
      catch( Exception e ) {
e.printStackTrace();
         throw new BundleException( "Unable to install bundle: " + location, e );
      }
   }

   protected void bundleUninstalled( long id ) throws BundleException {
      // don't remove bundle id
      // leave bundle present on framework in uninstalled state
      // (to completely remove all trace of the bundle from the framework, either stop and restart
      // framework, or call this class's wipeUninstalledBundles() method)
      // id_bundle.remove( new Long( id ) );

      // no need to unresolve bundles because it is done in BundleImpl's uninitializeBundle() method
      // unresolveBundles();

      log( LOG_INFO, "Bundle " + id + " uninstalled" );
   }

   public void wipeUninstalledBundles( Bundle[] bundlesToCheck ) {
      if ( bundlesToCheck == null ) {
         bundlesToCheck = getBundles();
      }

      long bundleId;
      for ( int i = 0; i < bundlesToCheck.length; i++ ) {
         if ( bundlesToCheck[ i ].getState() == bundlesToCheck[ i ].UNINSTALLED ) {
            bundleId = bundlesToCheck[ i ].getBundleId();
            id_bundle.remove( new Long( bundleId ) );
            deleteBundleDirectory( bundleId );
         }
      }
   }

// *****************************************************************************

   protected String getProperty( String key ) {
      Util.checkNullParameter( key, getClass().getName(), "getProperty", "key" );

      if ( key.equals( Constants.FRAMEWORK_LANGUAGE ) ) {
         return Locale.getDefault().getLanguage();
      }
      else if ( key.equals( Constants.FRAMEWORK_OS_NAME ) ) {
         String osName = System.getProperty( "os.name" );
         if ( osName.equals( "OS/2" ) ) {
            return "OS2";
         }
         else if ( osName.equals( "procnto" ) ) {
            return "QNX";
         }
         else if ( osName.equals( "Win95" )
                   || osName.equals( "Windows 95" ) ) {
            return "Windows95";
         }
         else if ( osName.equals( "Win98" )
                   || osName.equals( "Windows 98" ) ) {
            return "Windows98";
         }
         else if ( osName.equals( "WinNT" )
                   || osName.equals( "Windows NT" ) ) {
            return "WindowsNT";
         }
         else if ( osName.equals( "WinCE" )
                   || osName.equals( "Windows CE" ) ) {
            return "WindowsCE";
         }
         else if ( osName.equals( "Win2000" )
                   || osName.equals( "Windows 2000" ) ) {
            return "Windows2000";
         }
         else if ( osName.equals( "Mac OS X" ) ) {
            return "MacOS";
         }
         else {
            return osName;
         }
      }
      else if ( key.equals( Constants.FRAMEWORK_OS_VERSION ) ) {
         return System.getProperty( "os.version" );
      }
      else if ( key.equals( Constants.FRAMEWORK_PROCESSOR ) ) {
         String processorName = System.getProperty( "os.arch" );
         if ( processorName.equals( "psc1k" ) ) {
            return "Ignite";
         }
         else if ( processorName.equals( "power" )
                   || processorName.equals( "ppc" ) ) {
            return "PowerPC";
         }
         else if ( processorName.equals( "pentium" )
                   || processorName.equals( "i386" )
                   || processorName.equals( "i486" )
                   || processorName.equals( "i586" )
                   || processorName.equals( "i686" ) ) {
            return "x86";
         }
         else {
            return processorName;
         }
      }
      else if ( key.equals( Constants.FRAMEWORK_VENDOR ) ) {
         return VENDOR;
      }
      else if ( key.equals( Constants.FRAMEWORK_VERSION ) ) {
         return VERSION;
      }
      else {
         return null;
      }
   }

   protected long generateNewBundleId() {
      // create a new bundle id (bundle id values must never be reused)

      // see if a bundle id file exists (in which we store the last bundle id generated)
      long newBundleId = 0;
      try {
         newBundleId = Long.valueOf( readFile( getLastBundleIdFile() ) ).longValue() + 1;
      }
      catch ( Exception ignore ) {
         // either the file is not there (we never generated a bundle id before) or it has
         // been corrupted.
         // In either case, do as if we never generated a bundle id before.

         // if largest bundle id = 34, return 1001
         // if largest bundle id = 254, return 1001
         // if largest bundle id = 1234, return 2001
         // if largest bundle id = 12455, return 13001
         long thousands = largestBundleId / 1000 + 1;
         newBundleId = thousands * 1000 + 1;
      }

      while ( id_bundle.get( new Long( newBundleId ) ) != null ) {
         newBundleId++;
      }

      largestBundleId = newBundleId;
      try {
         writeFile( getLastBundleIdFile(), "" + largestBundleId );
      }
      catch ( IOException ignore ) {}

      return newBundleId;
/*
      long id = System.currentTimeMillis();
      while ( id_bundle.get( new Long( id ) ) != null ) {
         try {
            Thread.sleep( 10 );
         }
         catch( InterruptedException ignore ) {}
         id = System.currentTimeMillis();
      }
      return id;
*/
   }

   public BundleImpl getBundle( String location ) {
      Enumeration bundles = id_bundle.elements();
      BundleImpl bundle;
      while ( bundles.hasMoreElements() ) {
         bundle = (BundleImpl)bundles.nextElement();
         if ( bundle.getLocation().equals( location ) ) {
            return bundle;
         }
      }
      throw new IllegalArgumentException( "Bundle with requested location does not exist: " + location );
   }

   public BundleImpl getBundle( long id ) {
      BundleImpl bundle = (BundleImpl)id_bundle.get( new Long( id ) );
      if ( bundle == null ) {
         throw new IllegalArgumentException( "Bundle with requested id does not exist: " + id );
      }
      return bundle;
   }

   public BundleImpl[] getBundles() {
      Number[] ids = new Number[ id_bundle.size() ];
      Enumeration enum = id_bundle.keys();
      int i = 0;
      while ( enum.hasMoreElements() ) {
         ids[ i++ ] = (Long)enum.nextElement();
      }
      ids = Util.sortNumbers( ids );

      BundleImpl[] bundles = new BundleImpl[ ids.length ];
      for ( i = 0; i < ids.length; i++ ) {
         bundles[ i ] = (BundleImpl)id_bundle.get( ids[ i ] );
      }
      return bundles;
   }

   public PackageAdminImpl getPackageAdmin() {
      return packageAdmin;
   }

   public PermissionAdmin getPermissionAdmin() {
      return null;
   }

   protected ServiceRegistry getServiceRegistry() {
      return serviceRegistry;
   }

   protected File createFreshBundleDirectory( long id ) {
      File bundleDirectory = generateBundleDirectory( id );

      // If directory already exists, remove all it contains
      if ( bundleDirectory.exists() ) {
         if ( !Util.recursiveFileDelete( bundleDirectory ) ) {
            log( LOG_WARNING, "Could not delete directory: " + bundleDirectory.getAbsolutePath() );
         }
      }

      // Create bundle directory, and return
      bundleDirectory.mkdirs();
      return bundleDirectory;
   }

   public File getBundleInnerFilesDirectory( long id ) {
      File bundleInnerFilesDirectory = (File)bundleInnerFileDirectories.get( new Long( id ) );
      if ( bundleInnerFilesDirectory == null ) {
         // this is the first time we get the directory
         bundleInnerFilesDirectory = new File( generateBundleDirectory( id ), bundleInnerFilesDirName + getLargestBundleJarIndex( id ) );
         bundleInnerFileDirectories.put( new Long( id ), bundleInnerFilesDirectory );
         if ( bundleInnerFilesDirectory.exists() ) {
            // clean up old stuff from previous engine runs
            Util.recursiveFileDelete( bundleInnerFilesDirectory );
         }
         bundleInnerFilesDirectory.mkdirs();
      }
      return bundleInnerFilesDirectory;
   }

   public File getBundleDataDirectory( long id ) {
      File dataDirectory = new File( generateBundleDirectory( id ), bundleDataDirName );
      if ( !dataDirectory.exists() ) {
         dataDirectory.mkdirs();
      }
      return dataDirectory;
   }

   protected int getLargestBundleJarIndex( long id ) {
      File bundleDir = generateBundleDirectory( id );
      String[] listing = bundleDir.list();
      String indexString;
      int index;
      int largestBundleJarIndex = -1;
      for ( int i = 0; i < listing.length; i++ ) {
         if ( listing[ i ].startsWith( bundleJarNameBase ) && listing[ i ].endsWith( bundleJarNameExtension ) ) {
            indexString = listing[ i ].substring( bundleJarNameBase.length(), listing[ i ].length() - bundleJarNameExtension.length() );
            try {
               index = Integer.valueOf( indexString ).intValue();
               if ( largestBundleJarIndex < index ) {
                  largestBundleJarIndex = index;
               }
            }
            catch( NumberFormatException ignore ) {}
         }
      }
      if ( largestBundleJarIndex > -1 ) {
         return largestBundleJarIndex;
      }
      throw new IllegalArgumentException( "No bundle jar file found for requested bundle id: " + id );
   }

   protected File getCurrentBundleJarFile( long id ) {
      int jarIndex = getLargestBundleJarIndex( id );
      return new File( generateBundleDirectory( id ), bundleJarNameBase + jarIndex + bundleJarNameExtension );
   }

   protected File getNewBundleJarFile( long id ) {
      int index;
      try {
         int jarIndex = getLargestBundleJarIndex( id );
         index = jarIndex + 1;
      }
      catch( IllegalArgumentException e ) {
         index = 0;
      }

      return new File( generateBundleDirectory( id ), bundleJarNameBase + index + bundleJarNameExtension );
   }

   // Called to delete jar file that was created during an unsuccessful bundle update.
   protected boolean deleteBundleJarFile( File file ) {
      if ( file.exists() ) {
         return file.delete();
      }
      return true;
   }

   protected boolean deleteBundleDirectory( long id ) {
      File bundleDir = generateBundleDirectory( id );
      return bundleDir.delete();
   }

   protected void setBundleUninstalled( long id ) throws IOException {
      File uninstalledFlagFile = generateBundleUninstalledFlagFile( id );
      writeFile( uninstalledFlagFile, "" );
   }

   protected void registerBundleAsStarted( long bundleId ) throws IOException {
      File startedFlagFile = generateBundleStartedFlagFile( bundleId );
      writeFile( startedFlagFile, "" );
   }

   protected void recordBundleLocation( long bundleId, String location ) throws IOException {
      File file = generateBundleLocationFile( bundleId );
      writeFile( file, location );
   }

   protected String readStoredBundleLocation( long id ) throws IOException {
      File file = generateBundleLocationFile( id );
      if ( !file.exists() ) {
         throw new IllegalArgumentException( "No location file found for bundle: " + id );
      }

      return readFile( file );
   }

   protected boolean unregisterBundleAsStarted( Bundle bundle ) {
      File startedFlagFile = generateBundleStartedFlagFile( bundle.getBundleId() );
      if ( !startedFlagFile.exists() ) {
         return true;
      }
      else {
         return startedFlagFile.delete();
      }
   }

   protected int getReasonableTimeToWait() {
      return reasonableTimeToWait;
   }

// *****************************************************************************
// Security
// *****************************************************************************

   public abstract void checkAdminPermission() throws SecurityException;
   public abstract void checkFilePermission( String file, String permissionToCheck ) throws SecurityException;
   public abstract void checkPackagePermission( NamedPropertySet[] packages, String permissionToCheck, Object codeSource ) throws SecurityException;
   public abstract void checkPropertyPermission( String property, String permissionToCheck ) throws SecurityException;
   public abstract void checkServicePermission( String className, String permissionToCheck ) throws SecurityException;
   // Throw a security exception if the check fails on one ore more of the classes
   public abstract void checkServicePermissionForAll( String[] classNames, String permissionToCheck ) throws SecurityException;
   // Throw a security exception if all the checks fail
   public abstract void checkServicePermissionForAtLeastOne( String[] classNames, String permissionToCheck ) throws SecurityException;
   protected abstract boolean hasPermission( BundleImpl bundle, Object permission );
   protected abstract BundleClassLoader createClassLoaderForBundle( BundleImpl bundle, File bundleJarFile, NamedPropertySet[] nativeLibraries, NamedPropertySet[] classPath ) throws IOException;


// *****************************************************************************
// Events
// *****************************************************************************

   protected void addServiceListener( BundleImpl bundle, ServiceListener serviceListener, String filterString ) throws InvalidSyntaxException {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " adding a service listener" );
      if ( filterString == null ) {
         addServiceListener( bundle, serviceListener );
      }
      else {
         // use filter that use case-insensitive keys (service property names are case-insensitive)
         serviceListeners.addListener( bundle, serviceListener, new LDAPFilter( filterString, false ) );
      }
   }

   protected void addServiceListener( BundleImpl bundle, ServiceListener serviceListener ) {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " adding a service listener" );
      serviceListeners.addListener( bundle, serviceListener );
   }

   protected void removeServiceListener( BundleImpl bundle, ServiceListener serviceListener ) {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " removing a service listener" );
      serviceListeners.removeListener( bundle, serviceListener );
   }

   protected void removeServiceListenersRegisteredBy( BundleImpl bundle ) {
      int n = serviceListeners.getListenersRegisteredBy( bundle ).length;
      if ( n > 0 ) {
         log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " removing " + n + " service listener(s)" );
      }
      serviceListeners.removeAllListenersRegisteredBy( bundle );
   }

   protected void addBundleListener( BundleImpl bundle, BundleListener bundleListener ) {
      if ( bundleListener instanceof SynchronousBundleListener ) {
         // adding a synchronous bundle listener requires admin permission
         checkAdminPermission();
      }
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " adding a bundle listener" );
      bundleListeners.addListener( bundle, bundleListener );
   }

   protected void removeBundleListener( BundleImpl bundle, BundleListener bundleListener ) {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " removing a bundle listener" );
      bundleListeners.removeListener( bundle, bundleListener );
   }

   protected void removeBundleListenersRegisteredBy( BundleImpl bundle ) {
      int n = bundleListeners.getListenersRegisteredBy( bundle ).length;
      if ( n > 0 ) {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " removing " + n + " bundle listener(s)" );
      }
      bundleListeners.removeAllListenersRegisteredBy( bundle );
   }

   protected void addFrameworkListener( BundleImpl bundle, FrameworkListener frameworkListener ) {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " adding a framework listener" );
      frameworkListeners.addListener( bundle, frameworkListener );
   }

   protected void removeFrameworkListener( BundleImpl bundle, FrameworkListener frameworkListener ) {
      log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " removing a framework listeners" );
      frameworkListeners.removeListener( bundle, frameworkListener );
   }

   protected void removeFrameworkListenersRegisteredBy( BundleImpl bundle ) {
      int n = frameworkListeners.getListenersRegisteredBy( bundle ).length;
      if ( n > 0 ) {
         log( LOG_DEBUG, bundle.getLocationWithoutAdminPermissionCheck() + " removing " + n + " framework listener(s)" );
      }
      frameworkListeners.removeAllListenersRegisteredBy( bundle );
   }

   protected void fireServiceEvent( int type, ServiceReference serviceReference ) {
      if ( handlingEvents ) {
         final ServiceEvent event = new ServiceEvent( type, serviceReference );
         String[] serviceObjectClasses = (String[])serviceReference.getProperty( Constants.OBJECTCLASS );
         ServicePermission[] permissions = new ServicePermission[ serviceObjectClasses.length ];
         for ( int i = 0; i < serviceObjectClasses.length; i++ ) {
            permissions[ i ] = new ServicePermission( serviceObjectClasses[ i ], ServicePermission.GET );
         }

         Enumeration bundles = serviceListeners.getBundles();
         BundleImpl bundle;
         boolean permissionOk;
         Object[] bundleListeners;
         Vector listeners = new Vector();
         while ( bundles.hasMoreElements() ) {
            bundle = (BundleImpl)bundles.nextElement();
            // check if bundle has GET service permission for at least one of the classes
            // under which the service was registered
            permissionOk = false;
            for ( int i = 0; i < permissions.length; i++ ) {
               if ( hasPermission( bundle, permissions[ i ] ) ) {
                  permissionOk = true;
                  break;
               }
            }
            if ( permissionOk ) {
               // Only notify listeners if their bundle has GET service permission for
               // at least one of the classes under which the service was registered
               bundleListeners = serviceListeners.getListenersRegisteredBy( bundle );
               for ( int i = 0; i < bundleListeners.length; i++ ) {
                  LDAPFilter filter = serviceListeners.getFilter( bundleListeners[ i ] );
                  if ( filter == null || filter.match( serviceReference ) ) {
                     // Only notify listener if the filter it registered with is passed by the service
                     listeners.addElement( bundleListeners[ i ] );
                  }
               }
            }
         }

         // synchronous event delivery
         // OSGi 2.0  spec says service event delivery should occur synchronously
         // (i.e. in the same thread that caused the event to be fired
         for ( int i = 0; i < listeners.size(); i++ ) {
            try {
               doServiceEventCallback( (ServiceListener)listeners.elementAt( i ), event );
            }
            catch ( Throwable e ) {
               log( LOG_WARNING, "A service listener threw an exception while being notified of an event" );
            }
         }
      }
   }

   protected void fireBundleEvent( int type, Bundle bundle ) {
      if ( handlingEvents ) {
         SynchronousBundleListener[] synchronousTargets = bundleListeners.getSynchronousBundleListeners();
         final BundleListener[] nonSynchronousTargets = bundleListeners.getNonSynchronousBundleListeners();
         final BundleEvent event = new BundleEvent( type, bundle );
         // synchronous event delivery to synchronous listeners
         for ( int i = 0; i < synchronousTargets.length; i++ ) {
            try {
               doBundleEventCallback( synchronousTargets[ i ], event );
            }
            catch ( Throwable e ) {
               log( LOG_WARNING, "A bundle listener threw an exception while being notified of an event" );
            }
         }
         // asynchronous event delivery to other listeners
         new Thread( ( new Runnable() {
            public void run() {
               for ( int i = 0; i < nonSynchronousTargets.length; i++ ) {
                  try {
                     doBundleEventCallback( nonSynchronousTargets[ i ], event );
                  }
                  catch ( Throwable e ) {
                     log( LOG_WARNING, "A bundle listener threw an exception while being notified of an event" );
                  }
               }
            }
         } ) ).start();
      }
   }

   public void fireFrameworkEvent( int type, Bundle bundle, Throwable throwable ) {
      if ( type == FrameworkEvent.ERROR ) {
         log( LOG_ERROR, "Framework error" );
         if ( bundle != null ) {
            log( LOG_ERROR, "Bundle: " + bundle.getLocation() + " (Id: " + bundle.getBundleId() + ")" );
         }
         if ( throwable != null ) {
            StringBuffer sb = new StringBuffer( "Exception: " + throwable.toString() );
            if ( throwable instanceof BundleException ) {
               Throwable nested = ( (BundleException)throwable ).getNestedException();
               if ( nested != null ) {
                  sb.append( "\nNested excpetion: " + nested.toString() );
               }
            }
            log( LOG_ERROR, sb.toString() );
         }
      }

      if ( handlingEvents ) {
         final Enumeration targets = frameworkListeners.getAllListeners();
         final FrameworkEvent event = new FrameworkEvent( type, bundle, throwable );
         // asynchronous event delivery to all listeners
         new Thread( ( new Runnable() {
            public void run() {
               while ( targets.hasMoreElements() ) {
                  try {
                     doFrameworkEventCallback( (FrameworkListener)targets.nextElement(), event );
                  }
                  catch ( Throwable e ) {
                     log( LOG_WARNING, "A framework listener threw an exception while being notified of an event" );
                  }
               }
            }
         } ) ).start();
      }
   }

   protected void fireFrameworkErrorEvent( Bundle bundle, Throwable exception ) {
      log( LOG_ERROR, "Framework error" );
      log( LOG_ERROR, "Exception: " + exception.toString() );

      if ( handlingEvents ) {
         final Enumeration targets = frameworkListeners.getAllListeners();
         final FrameworkEvent event = new FrameworkEvent( FrameworkEvent.ERROR, bundle, exception );
         // asynchronous event delivery
         new Thread( ( new Runnable() {
            public void run() {
               while ( targets.hasMoreElements() ) {
                  try {
                     doFrameworkEventCallback( (FrameworkListener)targets.nextElement(), event );
                  }
                  catch ( Throwable e ) {
                     log( LOG_WARNING, "A framework listener threw an exception while being notified of an event" );
                  }
               }
            }
         } ) ).start();
      }
   }

   private void fireFrameworkStartedEvent() {
      if ( handlingEvents ) {
         final Enumeration targets = frameworkListeners.getAllListeners();
         final FrameworkEvent event = new FrameworkEvent( FrameworkEvent.STARTED, getBundle( 0 ), null );
         // asynchronous event delivery
         new Thread( ( new Runnable() {
            public void run() {
               while ( targets.hasMoreElements() ) {
                  try {
                     doFrameworkEventCallback( (FrameworkListener)targets.nextElement(), event );
                  }
                  catch ( Throwable e ) {
                     log( LOG_WARNING, "A framework listener threw an exception while being notified of an event" );
                  }
               }
            }
         } ) ).start();
      }
   }

   public void doServiceEventCallback( ServiceListener listener, ServiceEvent event ) {
      listener.serviceChanged( event );
   }

   public void doBundleEventCallback( BundleListener listener, BundleEvent event ) {
      listener.bundleChanged( event );
   }

   public void doFrameworkEventCallback( FrameworkListener listener, FrameworkEvent event ) {
      listener.frameworkEvent( event );
   }

// *****************************************************************************
// Low level framework logging
// *****************************************************************************

   public static final String LOG_INFO = "INFO: ";
   public static final String LOG_DEBUG = "DEBUG: ";
   public static final String LOG_WARNING = "WARNING: ";
   public static final String LOG_ERROR = "ERROR: ";

   private Hashtable logTypes = new Hashtable();

   private void initLog() {
      // log errors by default
      if ( ( "" + false ).equals( System.getProperty( "com.opensugar.cube.lowLevelLog.showErrors" ) ) ) {
         setLogging( LOG_ERROR, false );
      }
      else {
         setLogging( LOG_ERROR, true );
      }
      // log warnings by default
      if ( ( "" + false ).equals( System.getProperty( "com.opensugar.cube.lowLevelLog.showWarnings" ) ) ) {
         setLogging( LOG_WARNING, false );
      }
      else {
         setLogging( LOG_WARNING, true );
      }
      // do not log infos by default
      if ( ( "" + true ).equals( System.getProperty( "com.opensugar.cube.lowLevelLog.showInfos" ) ) ) {
         setLogging( LOG_INFO, true );
      }
      else {
         setLogging( LOG_INFO, false );
      }
   }

   protected void setLogging( String logType, boolean logging ) {
      if ( logging ) {
         logTypes.put( logType, "" );
      }
      else {
         logTypes.remove( logType );
      }
   }

   public void log( String logType, String message ) {
      log( logType, message, null );
   }

   public void log( String logType, String message, Throwable exception ) {
      if ( logTypes.get( logType ) != null ) {
         System.out.println( logType + message );
         if ( logType.equals( LOG_ERROR ) && exception != null ) {
            if ( exception instanceof BundleException && ( (BundleException)exception ).getNestedException() != null ) {
               ( (BundleException) exception ).getNestedException().printStackTrace();
            }
            else {
               exception.printStackTrace();
            }
         }
      }
   }












// *****************************************************************************
// Files and directories
// *****************************************************************************

   protected File getBaseDirectory() {
      if ( baseDirectory == null ) {
         baseDirectory = new File( System.getProperty( SYSTEM_PROPERTY_CUBE_DIR, defaultCubeDir ) );
      }
      return baseDirectory;
   }

   private File getLastBundleIdFile() {
      return new File( getBaseDirectory(), bundleIdFileName );
   }

   private File generateBundleDirectory( long id ) {
      return new File( getBaseDirectory(), bundleDirNameBase + id );
   }

   private File generateBundleStartedFlagFile( long id ) {
      return new File( generateBundleDirectory( id ), bundleStartedFlagFileName );
   }

   private File generateBundleUninstalledFlagFile( long id ) {
      return new File( generateBundleDirectory( id ), bundleUninstalledFlagFileName );
   }

   private File generateBundleLocationFile( long id ) {
      return new File( generateBundleDirectory( id ), bundleLocationFileName );
   }

   private String readFile( File file ) throws IOException {
      FileInputStream fis = null;
      try {
         fis = new FileInputStream( file );
         int n = fis.available();
         byte[] data = new byte[ n ];
         fis.read( data );
         return new String( data );
      }
      catch( IOException e ) {
         throw e;
      }
      finally {
         try {
            fis.close();
         }
         catch( Exception ignore ) {}
      }
   }

   private void writeFile( File file, String contents ) throws IOException {
      FileOutputStream fos = null;
      try {
         fos = new FileOutputStream( file );
         fos.write( contents.getBytes() );
      }
      catch( IOException e ) {
         throw e;
      }
      finally {
         try {
            fos.close();
         }
         catch( Exception ignore ) {}
      }
   }

// *****************************************************************************
// deleteUninstalledBundleDirectories()
// deleteObsoleteBundleJars()
// reinstallBundles()
// restartBundles()
// *****************************************************************************

   // When a bundle is uninstalled, it may not be possible to delete the bundle's
   // directory.  That is the case if the bundle's stop method does not release all
   // the files that it is using.  (I think the bundle is badly coded if it does that
   // but we want to be robust against this sort of behavior.)
   // If a bundle is uninstalled and its directory is not deleted, the engine records
   // the fact that this directory should have been deleted by creating an uninstalled
   // flag file in the directory.  This way, next time the engine is started it can
   // delete the directories of uninstalled bundles.
   // Here we check all the bundle directories, and delete any directory that contains
   // an uninstalled flag file.
   private void deleteUninstalledBundleDirectories() {
      File[] bundleDirs = getBundleDirectories();
      if ( bundleDirs == null ) {
         return;
      }

      for ( int i = 0; i < bundleDirs.length; i++ ) {
         if ( !isInstalled( bundleDirs[ i ] ) ) {
            if ( !Util.recursiveFileDelete( bundleDirs[ i ] ) ) {
               log( LOG_WARNING, "Unable to delete bundle directory: " + bundleDirs[ i ] );
            }
         }
      }
   }

   // Check if the specified directory contains an uninstalled flag file.
   private boolean isInstalled( File bundleDir ) {
      File uninstalledFlagFile = new File( bundleDir, AbstractCube.bundleUninstalledFlagFileName );
      if ( uninstalledFlagFile.exists() ) {
         return false;
      }
      return true;
   }

   // Check if the specified directory contains a started flag file.
   private boolean isStarted( File bundleDir ) {
      File startedFlagFile = new File( bundleDir, AbstractCube.bundleStartedFlagFileName );
      if ( startedFlagFile.exists() ) {
         return true;
      }
      return false;
   }

   // When a bundle is updated, it may not be possible to delete the old version of
   // the bundle's jar file (and the directory in which inner files are stored).
   // That is the case if the bundle's stop method does not
   // release all the files that it is using.  (I think the bundle is badly coded if it
   // does that but we want to be robust against this sort of behavior.)
   // That is also be the case if the cube needs to continue exporting classes that are
   // in the old bundle jar file but not in the new one.  (Note that if the cube continues
   // to export the packages of a bundle after it is updated or uninstalled, it must stop
   // doing so next time it is started.  Therefore deletion of the bundle jar here is the
   // correct behavior.)
   // If a bundle is updated and its old jar file is not deleted, the cube stores the
   // new jar file next to the old jar file, with an index number one higher than the index
   // of the old jar file.  When the engine is restarted it should delete the obsolete jar
   // files, i.e. all jar files in the bundle directory except that with the highest index
   // number.
   // Also, when a problem occurs during installation of a bundle, a directory for the bundle
   // can be created that does not contain the required files.  If directories exist that do
   // not contain all required files, they should be deleted so the engine does not try to
   // restart the bundle.
   //
   // Here we check all the bundle directories, and delete any jar file and inner file directory
   // with index number smaller than the index number of the jar file that has the highest index
   // number in the same directory.
   //
   // (If bundle jar name base is "bundle" and bundle jar name extension is .jar, names of jar
   // files take the form: bundle0.jar, bundle1.jar, bundle2.jar, and if the bundle inner file
   // directory base is "inner", then inner file directories take the form: inner0, inner1,
   // inner2, etc.
   private void deleteObsoleteBundleFiles() {
      File[] invalidBundleDirs = getInvalidBundleDirectories();
      for ( int i = 0; i < invalidBundleDirs.length; i++ ) {
         if ( !invalidBundleDirs[ i ].delete() ) {
            log( LOG_ERROR, "Unable to delete invalid bundle directory: " + invalidBundleDirs[ i ] );
            System.exit( 1 );
         }
      }


      File[] bundleDirs = getBundleDirectories();
      if ( bundleDirs == null ) {
         return;
      }

      File[] obsoleteBundleFiles;
      for ( int i = 0; i < bundleDirs.length; i++ ) {
         obsoleteBundleFiles = getObsoleteBundleFiles( bundleDirs[ i ] );
         if ( obsoleteBundleFiles == null ) {
            continue;
         }
         for ( int j = 0; j < obsoleteBundleFiles.length; j++ ) {
            if ( !Util.recursiveFileDelete( obsoleteBundleFiles[ j ] ) ) {
               log( LOG_WARNING, "Unable to delete obsolete bundle file: " + obsoleteBundleFiles[ j ].getPath() );
            }
         }
      }
   }

   // Return a list of all invalid bundle directories.
   // Invalid bundle directories are bundle directories that do not contain the files
   // that are required in a bundle directory.
   // The files that are required in a bundle directory are the bundle jar file and the
   // location file.
   private File[] getInvalidBundleDirectories() {
      File[] bundleDirs = getBundleDirectories();
      if ( bundleDirs == null ) {
         return new File[ 0 ];
      }
      String[] listing;
      Vector v = new Vector();
      for ( int i = 0; i < bundleDirs.length; i++ ) {
         listing = bundleDirs[ i ].list();
         boolean bundleJarFileFound = false;
         boolean locationFileFound = false;
         for ( int j = 0; j < listing.length; j++ ) {
            try {
               getBundleJarFileIndex( listing[ j ] );
               bundleJarFileFound = true;
            }
            catch( Exception e ) {
               // not a bundle directory
            }

            if ( listing[ j ].equals( AbstractCube.bundleLocationFileName ) ) {
               locationFileFound = true;
            }
         }
         if ( !bundleJarFileFound || !locationFileFound ) {
            v.addElement( bundleDirs[ i ] );
         }
      }

      File[] invalidBundleDirs = new File[ v.size() ];
      v.copyInto( invalidBundleDirs );
      return invalidBundleDirs;
   }

   // Return a list of all the bundle directories.
   // Bundle directories are direct subdirectories of the engine directory.
   // If the bundle directory name base is "b", the name of the bundle directories
   // take the form b1, b2, b3, etc, where the integer after the base name is the
   // bundle id.
   private File[] getBundleDirectories() {
      File baseDir = getBaseDirectory();
      if ( !baseDir.exists() ) {
         return null;
      }

      String[] listing = baseDir.list();
      Vector vec = new Vector();
      String bundleId;
      for ( int i = 0; i < listing.length; i++ ) {
         try {
            // Check if directory is a bundle directory
            getBundleId( listing[ i ] );
            // If so add it to the list.
            vec.addElement( new File( baseDir, listing[ i ] ) );
         }
         catch( IllegalArgumentException ignore ) {
            // Directory is not a bundle directory
         }
      }

      File[] bundleDirs = new File[ vec.size() ];
      vec.copyInto( bundleDirs );
      return bundleDirs;
   }


   // Return the bundle id corresponding to the specified bundle directory.
   // If the bundle directory name base is "b", the name of the bundle directories
   // take the form b1, b2, b3, etc, where the integer after the base name is the
   // bundle id.
   private long getBundleId( String bundleDirName ) throws IllegalArgumentException {
      if ( !bundleDirName.startsWith( AbstractCube.bundleDirNameBase ) ) {
         throw new IllegalArgumentException( "Directory name is not a valid bundle directory name" );
      }
      String idString = bundleDirName.substring( AbstractCube.bundleDirNameBase.length(), bundleDirName.length() );
      try {
         Long id = Long.valueOf( idString );
         return id.longValue();
      }
      catch( NumberFormatException e ) {
         throw new IllegalArgumentException( "Directory name is not a valid bundle directory name" );
      }
   }

   // Return a list of all the old (obsolete) bundle files in a given directory.
   // Obsolete bundle files are obsolete bundle jar files and obsolete bundle inner jar file
   // directories.
   // (See comments of method deleteObsoleteBundleJars()
   //
   // (If bundle jar name base is "bundle" and bundle jar name extension is .jar, names of jar
   // files take the form: bundle0.jar, bundle1.jar, bundle2.jar, etc
   // If bundle inner jar file directory name is "inner", names of inner jar file directories
   // take the form: inner0, inner1, inner2, etc
   private File[] getObsoleteBundleFiles( File bundleDir ) {
      if ( bundleDir == null ) {
         return null;
      }

      // Collect index number of all jar files.
      String[] listing = bundleDir.list();
      if ( listing == null ) {
         // bundleDir is in some weird state
         return null;
      }

      Vector indices = new Vector();
      for ( int i = 0; i < listing.length; i++ ) {
         try {
            indices.addElement( new Integer( getBundleJarFileIndex( listing[ i ] ) ) );
         }
         catch( Exception e ) {
            // not a bundle jar file
         }
      }

      // If no index has been collected, no jar file is in this directory.
      if ( indices.size() == 0 ) {
         log( LOG_WARNING, "No bundle jar file in directory: " + bundleDir );
         return null;
      }

      // If a single index has been collected, no obsolete jar file is in this directory.
      if ( indices.size() == 1 ) {
         return null;
      }

      // Find max bundle jar index.
      int maxIndex = -1;
      int index;
      for ( int i = 0; i < indices.size(); i++ ) {
         index = ( (Integer)indices.elementAt( i ) ).intValue();
         if ( maxIndex < index ) {
            maxIndex = index;
         }
      }

      // Remove max index from list to get list of obsolete jar indices.
      indices.removeElement( new Integer( maxIndex ) );

      // Create the list of bundle jar files and bundle inner jar file directories to return
      //
      // Assume that an inner file directory with a given index exists only if a bundle
      // jar file with that index exists.
      Vector tmp = new Vector();
      File innerFilesDirectory;
      for ( int i = 0; i < indices.size(); i++ ) {
         index = ( (Integer)indices.elementAt( i ) ).intValue();
         tmp.addElement( new File( bundleDir, AbstractCube.bundleJarNameBase + index + AbstractCube.bundleJarNameExtension ) );
         innerFilesDirectory = new File( bundleDir, AbstractCube.bundleInnerFilesDirName + index );
         if ( innerFilesDirectory.exists() ) {
            tmp.addElement( innerFilesDirectory );
         }
      }

      File[] obsoleteBundleFiles = new File[ tmp.size() ];
      tmp.copyInto( obsoleteBundleFiles );
      return obsoleteBundleFiles;
   }

   private int getBundleJarFileIndex( String bundleJarFileName ) throws NumberFormatException, IllegalArgumentException {
      if ( bundleJarFileName.startsWith( AbstractCube.bundleJarNameBase ) && bundleJarFileName.endsWith( AbstractCube.bundleJarNameExtension ) ) {
         String indexString = bundleJarFileName.substring( AbstractCube.bundleJarNameBase.length(), bundleJarFileName.length() - AbstractCube.bundleJarNameExtension.length()  );
         return Integer.valueOf( indexString ).intValue();
      }
      else {
         throw new IllegalArgumentException( "Invalid bundle jar file name" );
      }
   }

   // Install bundles that were installed last time the engine was run.
   //
   // OSGi 2 spec does not specify how to handle errors that may occur while
   // reinstalling bundles
   private void reinstallBundles() {
      File[] bundleDirs = getBundleDirectories();
      if ( bundleDirs == null ) {
         return;
      }

      long bundleId;
      for ( int i = 0; i < bundleDirs.length; i++ ) {
         if ( isInstalled( bundleDirs[ i ] ) ) {
            bundleId = getBundleId( bundleDirs[ i ].getName() );
            try {
               reinstallBundle( bundleId );
            }
            catch( BundleException e ) {
               log( LOG_ERROR, "Could not reinstall bundle " + bundleId, e );
            }
            catch( IOException e ) {
               log( LOG_ERROR, "Could not reinstall bundle " + bundleId, e );
            }
         }
      }
   }

   // Start bundles that were active last time the engine was run.
   //
   // OSGi 2 spec says that BundleException while restarting bundles should be
   // broadcast as framework error events
   private void restartBundles( Observer progressListener ) {
      File[] bundleDirs = getBundleDirectories();
      if ( bundleDirs == null ) {
         return;
      }

      // order bundles that need to be started in increasing bundle id order
      long bundleId, id;
      Vector vec = new Vector();
      for ( int i = 0; i < bundleDirs.length; i++ ) {
         if ( isStarted( bundleDirs[ i ] ) ) {
            bundleId = getBundleId( bundleDirs[ i ].getName() );
            int index = vec.size();
            for ( int j = 0; j < vec.size(); j++ ) {
               id = ( (Long)vec.elementAt( j ) ).longValue();
               if ( bundleId < id ) {
                  index = j;
                  break;
               }
            }
            vec.insertElementAt( new Long( bundleId ), index );
         }
      }

      int fudgeSleepTime = 0;
      try {
         fudgeSleepTime = Integer.valueOf( System.getProperty( "com.opensugar.cube.fudgeSleep" ) ).intValue();
      }
      catch ( Exception ignore ) {}

                        int percent = 0;
                        String percentString = percent + " %";
                        System.out.print( percentString );
                        if ( progressListener != null ) {
                           progressListener.update( null, new Integer( percent ) );
                        }
      for ( int i = 0; i < vec.size(); i++ ) {
         bundleId = ( (Long)vec.elementAt( i ) ).longValue();
         BundleImpl bundle = getBundle( bundleId );
         try {
                        for ( int j = 0; j < percentString.length(); j++ ) {
                           System.out.print( "\b" );
                        }
                        percent = (int)( ( i + 0.5 ) / vec.size() * 100 );
                        percentString = percent + " %";
                        System.out.print( percentString );
                        if ( progressListener != null ) {
                           progressListener.update( null, new Integer( percent ) );
                        }
            // check bundle is resolved before attempting to start it
            // (bundle could have been running last time the cube was shut down, using stale
            // packages exported by an updated or uninstalled bundle. If that is the case,
            // the stale packages are no longer available now that the cube is being restarted
            // - unless another bundle provides those packages), therefore the bundle cannot
            // be started
            if ( bundle.getState() == bundle.RESOLVED ) {
               bundle.start();
            }
            else {
               log( LOG_WARNING, "\nCould not restart bundle " + bundleId + " because it is not resolved" );
               PackageImport[] unsatisfiedImports = packageAdmin.getUnsatisfiedImports( bundle );
               System.out.println( "Missing packages:" );
               for ( int j = 0; j < unsatisfiedImports.length; j++ ) {
                  System.out.print( "\t" + unsatisfiedImports[ j ].getName() );
                  if ( unsatisfiedImports[ j ].getSpecificationVersion() != null ) {
                     System.out.print( " - v" + unsatisfiedImports[ j ].getSpecificationVersion() );
                  }
                  System.out.print( "\n" );
               }
            }
            if ( fudgeSleepTime > 0 ) {
               try {
                  Thread.sleep( fudgeSleepTime );
               }
               catch ( InterruptedException ignore ) {}
            }
                        for ( int j = 0; j < percentString.length(); j++ ) {
                           System.out.print( "\b" );
                        }
                        percent = (int)( ( i + 1.0 ) / vec.size() * 100 );
                        percentString = percent + " %";
                        System.out.print( percentString );
                        if ( progressListener != null ) {
                           progressListener.update( null, new Integer( percent ) );
                        }
         }
         catch( Throwable e ) {
            fireFrameworkEvent( FrameworkEvent.ERROR, bundle, e );
            log( LOG_ERROR, "\nCould not restart bundle " + bundleId, e );
e.printStackTrace();
         }
      }
                        for ( int j = 0; j < percentString.length(); j++ ) {
                           System.out.print( "\b" );
                        }
                        if ( progressListener != null ) {
                           progressListener.update( null, new Integer( percent ) );
                        }
   }

}
