/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube;

import com.opensugar.cube.ldap.LDAPFilter;
import com.opensugar.cube.serviceRegistry.ServiceRegistryException;

import org.osgi.framework.BundleContext;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.BundleListener;
import org.osgi.framework.FrameworkListener;
import org.osgi.framework.ServiceRegistration;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.ServiceFactory;
import org.osgi.framework.ServicePermission;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.Constants;
import org.osgi.framework.Filter;

import java.io.InputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Dictionary;
import java.util.Vector;
import java.util.Hashtable;
import java.net.URL;

public class BundleContextImpl implements BundleContext {

   public String getProperty( String key ) {
      Util.checkNullParameter( key, getClass().getName(), "getProperty", "key" );

      String value = cube.getProperty( key );
      if ( value != null ) {
         return value;
      }
      else {
         // if property is not a framework property
         // and if bundle has PropertyPermission "read",
         // then search the system properties for the property and return that
         try {
            cube.checkPropertyPermission( key, "read" );
            return (String)System.getProperties().get( key );
         }
         catch ( SecurityException e ) {
            return null;
         }
      }
   }

   public Bundle getBundle() {
      checkBundleContextIsNotStopped();

      return bundle;
   }

   public Bundle installBundle( String location ) throws BundleException, SecurityException {
      Util.checkNullParameter( location, getClass().getName(), "installBundle", "location" );

      return installBundle( location, null );
   }

   public Bundle installBundle( String location, InputStream in ) throws BundleException, SecurityException {
      Util.checkNullParameter( location, getClass().getName(), "installBundle", "location" );

      cube.checkAdminPermission();

      return cube.installBundle( location, in );
   }

   public Bundle getBundle( long id ) {
      return cube.getBundle( id );
   }

   public Bundle[] getBundles() {
      return cube.getBundles();
   }

   public Filter createFilter( String filter ) throws InvalidSyntaxException {
      if ( filter == null ) {
         throw new NullPointerException( "Filter string is null" );
      }
      // return filter that uses case-sensitive keys
      return new LDAPFilter( filter, false );
   }

   public void addServiceListener( ServiceListener listener, String filterString ) throws InvalidSyntaxException {
      Util.checkNullParameter( listener, getClass().getName(), "addServiceListener", "listener" );
      // filterString can be null

      checkBundleContextIsNotStopped();

      cube.addServiceListener( bundle, listener, filterString );
   }

   public void addServiceListener( ServiceListener listener ) {
      Util.checkNullParameter( listener, getClass().getName(), "addServiceListener", "listener" );

      checkBundleContextIsNotStopped();

      cube.addServiceListener( bundle, listener );
   }

   public void removeServiceListener( ServiceListener listener ) {
      Util.checkNullParameter( listener, getClass().getName(), "removeServiceListener", "listener" );

      checkBundleContextIsNotStopped();

      cube.removeServiceListener( bundle, listener );
   }

   public void addBundleListener( BundleListener listener ) {
      Util.checkNullParameter( listener, getClass().getName(), "addBundleListener", "listener" );

      checkBundleContextIsNotStopped();

      cube.addBundleListener( bundle, listener );
   }

   public void removeBundleListener( BundleListener listener ) {
      Util.checkNullParameter( listener, getClass().getName(), "removeBundleListener", "listener" );

      checkBundleContextIsNotStopped();

      cube.removeBundleListener( bundle, listener );
   }

   public void addFrameworkListener( FrameworkListener listener ) {
      Util.checkNullParameter( listener, getClass().getName(), "addFrameworkListener", "listener" );

      checkBundleContextIsNotStopped();

      cube.addFrameworkListener( bundle, listener );
   }

   public void removeFrameworkListener( FrameworkListener listener ) {
      Util.checkNullParameter( listener, getClass().getName(), "removeFrameworkListener", "listener" );

      checkBundleContextIsNotStopped();

      cube.removeFrameworkListener( bundle, listener );
   }

   public ServiceRegistration registerService( String[] classNames, Object service, Dictionary properties ) {
      Util.checkNullArrayParameter( classNames, getClass().getName(), "registerService", "classNames" );
      Util.checkNullParameter( service, getClass().getName(), "registerService", "service" );
      // properties can be null

      if ( properties == null ) {
         properties = new Hashtable();
      }
      properties.put( Constants.SERVICE_ID, generateServiceId() );
      properties.put( Constants.OBJECTCLASS, classNames );

      if ( !( service instanceof ServiceFactory ) ) {
         Class clazz;
         for ( int i = 0; i < classNames.length; i++ ) {
            try {
               clazz = bundle.getClassLoader().loadClass( classNames[ i ] );
               if ( !clazz.isInstance( service ) ) {
                  throw new IllegalArgumentException( "Error while registering service: " + service.getClass().getName() + " is not an instance of " + clazz.getName() );
               }
            }
            catch( ClassNotFoundException e ) {
               getCube().log( getCube().LOG_ERROR, "classNotFound: " + getBundle().getLocation(), e );
               throw new IllegalArgumentException( "Invalid class name: " + classNames[ i ] );
            }
         }
      }

      cube.checkServicePermissionForAll( classNames, ServicePermission.REGISTER );

      // osgi_compliance
      // framework lifecycle test does not like it when we do this check.
      // but the spec says we should do it
      //checkBundleContextIsNotStopped();

      ServiceRegistration serviceRegistration = cube.getServiceRegistry().registerService( bundle, classNames, service, properties );
      cube.fireServiceEvent( ServiceEvent.REGISTERED, serviceRegistration.getReference() );
      return serviceRegistration;
   }

   public ServiceRegistration registerService( String className, Object service, Dictionary properties ) {
      String[] classNames = { className };
      return registerService( classNames, service, properties );
   }
/*
   public ServiceReference[] getServiceReferences( String className, String filterString ) throws InvalidSyntaxException {
      // osgi_compliance
      // The OSGi 2.0 spec does not say we should throw an IllegalStateException if
      // the bundle context is stopped here.
      // But the framework lifecycle compliance test fails, if we don't!!!
      checkBundleContextIsNotStopped();

      return getServiceReferences( className, filterString, true );
   }
*/
//   private ServiceReference[] getServiceReferences( String className, String filterString, boolean checkBundleContextIsNotStopped ) throws InvalidSyntaxException {
   public ServiceReference[] getServiceReferences( String className, String filterString ) throws InvalidSyntaxException {
      // className can be null (in that case, return all registered services)
      // filterString can be null

      String[] classNames = { className };
      try {
         cube.checkServicePermissionForAll( classNames, ServicePermission.GET );
      }
      catch( SecurityException e ) {
         return null;
      }

//      if ( checkBundleContextIsNotStopped ) {
         checkBundleContextIsNotStopped();
//      }

      LDAPFilter filter = null;
      if ( filterString != null ) {
         // filter should use case-insensitive keys (service property names are case-insensitive)
         filter = new LDAPFilter( filterString, false );
      }
      return cube.getServiceRegistry().getServiceReferences( className, filter );
   }

   public ServiceReference getServiceReference( String className ) {
      // className can be null (in that case, return all registered services)

      String[] classNames = { className };
      try {
//         ServiceReference[] serviceReferences = getServiceReferences( className, null, true );
         ServiceReference[] serviceReferences = getServiceReferences( className, null );
         if ( serviceReferences != null && serviceReferences.length > 0 ) {
            // find services with largest ranking
            int largestRanking = Integer.MIN_VALUE;
            Object val;
            int ranking;
            for ( int i = 0; i < serviceReferences.length; i++ ) {
               val = serviceReferences[ i ].getProperty( Constants.SERVICE_RANKING );
               // val can be null, it can be an Integer, or it can be a String representing an integer
               if ( val != null ) {
                  ranking = Integer.valueOf( val.toString() ).intValue();
                  if ( ranking > largestRanking ) {
                     largestRanking = ranking;
                  }
               }
            }
            Vector largestRankingServices = new Vector();
            if ( largestRanking == Integer.MIN_VALUE ) {
               for ( int i = 0; i < serviceReferences.length; i++ ) {
                  largestRankingServices.addElement( serviceReferences[ i ] );
               }
            }
            else {
               for ( int i = 0; i < serviceReferences.length; i++ ) {
                  val = serviceReferences[ i ].getProperty( Constants.SERVICE_RANKING );
                  // val can be null, it can be an Integer, or it can be a String representing an integer
                  if ( val != null && Integer.valueOf( val.toString() ).intValue() == largestRanking ) {
                     largestRankingServices.addElement( serviceReferences[ i ] );
                  }
               }
            }

            // among the services that tie for highest ranking, return service with
            // smallest id (i.e. service that was registered first)
            ServiceReference largestRankingServiceWithSmallestId = null;
            ServiceReference sr;
            long smallestId = Long.MAX_VALUE;
            long id;
            for ( int i = 0; i < largestRankingServices.size(); i++ ) {
               sr = (ServiceReference)largestRankingServices.elementAt( i );
               id = ( (Long)serviceReferences[ i ].getProperty( Constants.SERVICE_ID ) ).longValue();
               if ( id < smallestId ) {
                  largestRankingServiceWithSmallestId = sr;
                  smallestId = id;
               }
            }
            return largestRankingServiceWithSmallestId;
         }
      }
      catch( InvalidSyntaxException e ) {
         // This cannot happen (the filter is null)
         throw new UnreachableCodeException();
      }
      return null;
   }

   public Object getService( ServiceReference reference ) {
      Util.checkNullParameter( reference, getClass().getName(), "getService", "reference" );

      cube.checkServicePermissionForAtLeastOne( (String[])reference.getProperty( Constants.OBJECTCLASS ), ServicePermission.GET );

      checkBundleContextIsNotStopped();

      try {
         Object service = cube.getServiceRegistry().getService( reference, bundle );
         return service;
      }
      catch( IllegalStateException e ) {
         // service has been unregistered
         return null;
      }
      catch( ServiceRegistryException e ) {
         // Either the service factory threw an exception while creating a new service instance,
         // or it created a service instance that was not an instance of all the classes named when
         // the service was registered, or some other error occured while creating a new service instance
         cube.fireFrameworkErrorEvent( getBundle(), e );
         return null;
      }
   }

   public boolean ungetService( ServiceReference reference ) {
      Util.checkNullParameter( reference, getClass().getName(), "ungetService", "reference" );

      checkBundleContextIsNotStopped();

      return cube.getServiceRegistry().ungetService( reference, bundle );
   }

   public File getDataFile( String fileName ) {
      Util.checkNullParameter( fileName, getClass().getName(), "getDataFile", "fileName" );

      checkBundleContextIsNotStopped();

      File dataDirectory = cube.getBundleDataDirectory( bundle.getBundleId() );
      cube.checkFilePermission( dataDirectory.getAbsolutePath() + File.separator + "-", "read,write,execute,delete" );
      if ( fileName.length() == 0 ) {
         return dataDirectory;
      }
      return new File( dataDirectory, fileName );
   }

// *****************************************************************************

   private static long lastAssignedServiceId = 0;

   private AbstractCube cube;
   private BundleImpl bundle;
   private boolean stopped;

   protected BundleContextImpl( AbstractCube cube, BundleImpl bundle ) {
      this.cube = cube;
      this.bundle = bundle;
      stopped = false;
   }

   public AbstractCube getCube() {
      return cube;
   }

   protected void setStopped() {
      stopped = true;
   }

// *****************************************************************************

   private void checkBundleContextIsNotStopped() {
      if ( stopped ) {
         throw new IllegalStateException( "Bundle context has been stopped (bundle id: " + bundle.getBundleId()  + ")" );
      }
   }

   private static synchronized Long generateServiceId() {
      long id = lastAssignedServiceId + 1;
      lastAssignedServiceId = id;
      return new Long( id );
   }

}