/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube;

import com.opensugar.cube.java2.ManifestHelper;
import com.opensugar.cube.packageAdmin.ExportedPackageImpl;
import com.opensugar.cube.packageAdmin.PackageImport;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.PackagePermission;
import org.osgi.framework.ServicePermission;
import org.osgi.framework.Constants;

import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Dictionary;
import java.util.Hashtable;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.zip.ZipFile;
import java.util.Vector;

public class BundleImpl implements Bundle {

   public int getState() {
      return state;
   }

   public void start() throws BundleException {
      getCube().checkAdminPermission();

      // If the bundle in uninstalled, an illegal state exception is thrown
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot start a bundle after it has been uninstalled" );
      }

      // If the bundle is starting or stopping, wait for the bundle to change state
      // before continuing.  If this does not occur in a reasonable time, throw a
      // bundle exception
      if ( getState() == STARTING || getState() == STOPPING ) {
         if ( !waitUntilStartedOrStopped() ) {
            throw new BundleException( "Cannot stop a bundle that is stuck in starting or stopping mode" );
         }
      }

      // If the bundle is active return immediately
      if ( getState() == ACTIVE ) {
         return;
      }

      // If the bundle is not resolved, attempt to resolve it.  If it cannot be
      // resolved, throw a bundle exception
      if ( getState() != RESOLVED ) {
         // This should have no effect because packages are processed every time
         // a bundle is installed or updated.
         // Do it anyway since the spec says so.
         getCube().getPackageAdmin().resolveBundles();
         if ( getState() != RESOLVED ) {
            PackageImport[] unsatisfiedImports = getCube().getPackageAdmin().getUnsatisfiedImports( this );

            StringBuffer sb = new StringBuffer();
            String version;
            for ( int i = 0; i < unsatisfiedImports.length; i++ ) {
               sb.append( unsatisfiedImports[ i ].getName() );
               version = unsatisfiedImports[ i ].getSpecificationVersion();
               if ( version != null ) {
                  sb.append( " " + version );
               }
               if ( i < unsatisfiedImports.length - 1 ) {
                  sb.append( ", " );
               }
            }
            throw new BundleException( "Cannot start a bundle that is not resolved.  Missing packages: " + sb.toString() );
         }
      }

      // Set the state to STARTING
      state = STARTING;

      // If a bundle activator is specified call its start method
      // If an error occurs, set the state back to RESOLVED, remove bundle listeners,
      // unregister services registered, release services used, and throw a bundle
      // exception
      bundleContext = new BundleContextImpl( cube, this );
      try {
         String bundleActivatorName = (String)headers.get( "Bundle-Activator" );
         if ( bundleActivatorName != null ) {
            Class bundleActivatorClass = classLoader.loadClass( bundleActivatorName.trim() );
            bundleActivator = (BundleActivator)bundleActivatorClass.newInstance();
            bundleActivator.start( bundleContext );
         }
      }
      catch( Throwable e ) {
         state = RESOLVED;
         bundleActivator = null;
         // Remove bundle listeners, unregister services registered, release services used
         cleanUp();
         throw new BundleException( "Exception occured in the bundle activator's start method", e );
      }

      // Record that this bundle has been started, so it gets restarted automatically
      // next time the framework is started.
      try {
         getCube().registerBundleAsStarted( getBundleId() );
      }
      catch( IOException e ) {
         getCube().fireFrameworkEvent( FrameworkEvent.ERROR, this, e );
      }

      // Set the state to ACTIVE
      state = ACTIVE;

      // Broadcast  bundle started event
      getCube().fireBundleEvent( BundleEvent.STARTED, this );
   }

   public void stop() throws BundleException {
      stop( false );
   }

   public void update() throws BundleException, SecurityException {
      update( null );
   }

   public void update( InputStream in ) throws BundleException, SecurityException {
      // in can be null

      getCube().checkAdminPermission();

      // If the bundle is uninstalled, an illegal state exception is thrown
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot update a bundle after it has been uninstalled" );
      }

      // Record whether or not this bundle is started, so we know if we need to
      // restart it after it has been updated
      boolean restart = false;
      if ( getState() == ACTIVE ) {
         restart = true;
      }

      // If the bundle is active or starting, it is stopped using the stop() method
      // If the stop() method throws an exception, the exception is rethrown and this
      // method exits
      if ( getState() == ACTIVE || getState() == STARTING ) {
         stop();
      }

      // Determine location of the version of bundle
      String tempLocation = (String)headers.get( "Bundle-UpdateLocation" );
      if ( tempLocation == null || tempLocation.trim().length() == 0 ) {
         tempLocation = location;
      }

      // Install new version of bundle
      // If an exception occurs while installing new version, restore original version
      Throwable exceptionWhileInstalling = null;
      String oldLocation = location;
      File oldBundleJarFile = bundleJarFile;
      File newBundleJarFile = getCube().getNewBundleJarFile( id );
      try {
         // Retreive new jar file and store in bundle directory
         // Close input stream even if exception is thrown
         if ( in == null ) {
            URL url = new URL( tempLocation );
            in = url.openStream();
         }

         Util.transferData( in, new FileOutputStream( newBundleJarFile ) );

         // Uninitialize bundle first so that class loader can be closed
         uninitializeBundle();

         // Re-initialize bundle
         initializeBundle( tempLocation, getCube().getCurrentBundleJarFile( id ) );
      }
      catch( Throwable e ) {
         exceptionWhileInstalling = e;
         // Restore original version (bundle has not been uninitialized, so old values still work)
         if ( getCube().deleteBundleJarFile( newBundleJarFile ) ) {
            try {
               initializeBundle( oldLocation, oldBundleJarFile );
            }
            catch( Throwable e2 ) {
               throw new BundleException( "Fatal bundle error: failure to restore original bundle version after unsuccessful bundle update attempt", e );
            }
         }
         else {
            throw new BundleException( "Fatal bundle error: failure to delete jar file that was created during unsuccessful bundle update attempt", e );
         }
      }

      // Set the state to INSTALLED
      state = INSTALLED;

      // If the new version was successfully installed, broadcast a bundle updated
      // event
      if ( exceptionWhileInstalling == null ) {
         getCube().fireBundleEvent( BundleEvent.UPDATED, this );
      }

      getCube().getPackageAdmin().resolveBundles();

      // If the bundle was originally active, start the updated bundle using the
      // start() method (the bundle will be resolved if it can in the start method)
      // If the start method throws an exception, broadcast a framework error event
      if ( restart ) {
         try {
            start();
         }
         catch( Throwable e ) {
            getCube().fireFrameworkErrorEvent( this, e );
         }
      }

//      getCube().unresolveBundles();
//      getCube().resolveBundles();

      // Throw a bundle exception if an error occured while attempting to install
      // new version of bundle
      if ( exceptionWhileInstalling != null ) {
         throw new BundleException( "Exception occured while attempting to install new version of bundle. Original version was restored.", exceptionWhileInstalling );
      }
   }

   public void uninstall() throws BundleException {
      getCube().checkAdminPermission();

      // If the bundle is uninstalled, an illegal state exception is thrown
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot uninstall bundle that is already uninstalled" );
      }

      // If the bundle is active or starting, stop it using the stop() method
      // If the stop() method throws an exception, broadcast framework error event
      if ( getState() == ACTIVE || getState() == STARTING ) {
         try {
            stop();
         }
         catch( Throwable e ) {
            getCube().fireFrameworkErrorEvent( this, e );
         }
      }

      // Uninitialize bundle
      try {
         uninitializeBundle();
      }
      catch( IOException e ) {
         getCube().fireFrameworkEvent( FrameworkEvent.ERROR, this, e );
      }

      try {
         getCube().setBundleUninstalled( id );
      }
      catch ( IOException e ) {
         getCube().fireFrameworkEvent( FrameworkEvent.ERROR, this, e );
      }
      getCube().bundleUninstalled( id );

      // Broadcast bundle uninstalled event
      getCube().fireBundleEvent( BundleEvent.UNINSTALLED, this );

      // Set the state to UNINSTALLED
      state = UNINSTALLED;
   }

   public Dictionary getHeaders() {
      getCube().checkAdminPermission();
      return headers;
   }

   public long getBundleId() {
      return id;
   }

   public String getLocation() {
      getCube().checkAdminPermission();
      return getLocationWithoutAdminPermissionCheck();
   }

   // used only for cube to do debug logging, so that bundle location can be included
   // in operation logs, even if the code being executed does not have AdminPermission,
   // which is normally required to get the bundle location;
   public String getLocationWithoutAdminPermissionCheck() {
      return location.trim();
   }

   public ServiceReference[] getRegisteredServices() {
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot get services registered by a bundle after it has been uninstalled" );
      }

      ServiceRegistrationImpl[] serviceRegistrations = getCube().getServiceRegistry().getServicesRegisteredBy( this );
      if ( serviceRegistrations == null || serviceRegistrations.length == 0 ) {
         return null;
      }

      Vector tmp = new Vector();
      ServiceReference serviceReference;
      for ( int i = 0; i < serviceRegistrations.length; i++ ) {
         serviceReference = serviceRegistrations[ i ].getReference();
         try {
            getCube().checkServicePermissionForAtLeastOne( (String[])serviceReference.getProperty( "objectClass" ), ServicePermission.GET );
            tmp.addElement( serviceReference );
         }
         catch( SecurityException ignore ) {}
      }

      if ( tmp.size() == 0 ) {
         return null;
      }
      ServiceReference[] ret = new ServiceReference[ tmp.size() ];
      tmp.copyInto( ret );
      return ret;
   }

   public ServiceReference[] getServicesInUse() {
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot get services used by a bundle after it has been uninstalled" );
      }

      ServiceReference[] serviceReferences = getCube().getServiceRegistry().getServicesUsedBy( this );
      if ( serviceReferences == null || serviceReferences.length == 0 ) {
         return null;
      }

      Vector tmp = new Vector();
      for ( int i = 0; i < serviceReferences.length; i++ ) {
         try {
            getCube().checkServicePermissionForAtLeastOne( (String[])serviceReferences[ i ].getProperty( "objectClass" ), ServicePermission.GET );
            tmp.addElement( serviceReferences[ i ] );
         }
         catch( SecurityException ignore ) {}
      }

      if ( tmp.size() == 0 ) {
         return null;
      }
      ServiceReference[] ret = new ServiceReference[ tmp.size() ];
      tmp.copyInto( ret );
      return ret;
   }

   public boolean hasPermission( Object permission ) {
      Util.checkNullParameter( permission, getClass().getName(), "hasPermission", "permission" );

      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot check a bundle's permissions after it has been uninstalled" );
      }

      return getCube().hasPermission( this, permission );
   }

   public int compareTo( Object obj ) {
      Util.checkNullParameter( obj, getClass().getName(), "compareTo", "obj" );

      if ( obj instanceof Bundle ) {
         Bundle bundleToCompareTo = (Bundle)obj;
         return (int)( getBundleId() - bundleToCompareTo.getBundleId() );
      }
      throw new IllegalArgumentException( "Object to compare to must be of type org.osgi.framework.Bundle" );
   }

   public URL getResource( String name ) throws IllegalStateException {
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot get resource after bundle has been uninstalled" );
      }

      try {
         getCube().checkAdminPermission();
      }
      catch( SecurityException e ) {
         return null;
      }

/*
// ???
// look for resource using current class loader
// if not found using current class loader, try getting resource using old class loaders
URL url = classLoader.getResource( name );
if ( oldClassLoaders != null ) {
   int i = 0;
   while( url == null && i < oldClassLoaders.size() ) {
      url = ( (BundleClassLoader)oldClassLoaders.elementAt( i ) ).getResource( name );
   }
}
return url;
*/

      // shared packages should only be searched if this bundle is resolved
      // this is ok because when classloader tries to search shared packages, package admin
      // will always return null if the bundle is not resolved
      return classLoader.getResource( name );
   }

   public NamedPropertySet[] getExportedPackages() {
      return exportedPackages;
   }

   public NamedPropertySet[] getExplicitlyImportedPackages() {
      return explicitlyImportedPackages;
   }

// *****************************************************************************

   // Independant of bundle updates
   private AbstractCube cube;
   private long id;

   // Change when bundle is updated
   private String location;
   private File bundleJarFile;
   private Dictionary headers;
   private NamedPropertySet[] exportedPackages;
   private NamedPropertySet[] explicitlyImportedPackages;
   private BundleClassLoader classLoader;

   // changes when the bundle is stopped
   private BundleContextImpl bundleContext;

   private int state;
   private BundleActivator bundleActivator;

   protected BundleImpl( AbstractCube cube, long id, String location, File bundleJarFile ) throws IOException, ClassNotFoundException, IllegalAccessException, InstantiationException {
      this.cube = cube;
      this.id = id;

      // initialize bundle context (so that things can be done before the bundle
      // is started - getDataFile() for instance)
      bundleContext = new BundleContextImpl( cube, this );

      initializeBundle( location, bundleJarFile );
   }

   public BundleContextImpl getBundleContext() {
      return bundleContext;
   }

   // Called by package admin
   public void setResolved() {
      if ( getState() != INSTALLED ) {
         throw new RuntimeException( "Cannot resolve a bundle that is not in the installed state" );
      }

      state = RESOLVED;
   }

   // Called by package admin
   public void setInstalled() {
      if ( getState() != RESOLVED ) {
         throw new RuntimeException( "Cannot unresolve a bundle that is not in the resolved state" );
      }
      state = INSTALLED;

      // Broadcast  bundle installed event
      getCube().fireBundleEvent( BundleEvent.INSTALLED, this );
   }

   public BundleClassLoader getClassLoader() {
      return classLoader;
   }

   public AbstractCube getCube() {
      return cube;
   }

   public void stop( boolean keepRegisteredAsStarted ) throws BundleException {
      getCube().checkAdminPermission();

      // If the bundle is uninstalled, an illegal state exception is thrown
      if ( getState() == UNINSTALLED ) {
         throw new IllegalStateException( "Cannot stop a bundle after it has been uninstalled" );
      }

      // If the bundle is starting or stopping, wait for the bundle to change state
      // before continuing.  If this does not occur in a reasonable time, throw a
      // bundle exception
      if ( getState() == STARTING || getState() == STOPPING ) {
         if ( !waitUntilStartedOrStopped() ) {
            throw new BundleException( "Cannot stop a bundle that is stuck in starting or stopping mode" );
         }
      }

      // If the bundle is not active return immediately
      if ( getState() != ACTIVE ) {
         return;
      }

      // Set the state to STOPPING
      state = STOPPING;

      // If keepRegisteredAsStarted flag is false, record that this bundle has been
      // stopped, so it does not restart automatically next time the framework is started.
      if ( !keepRegisteredAsStarted ) {
         if ( !getCube().unregisterBundleAsStarted( this ) ) {
            getCube().fireFrameworkErrorEvent( this, new Exception( "Bundle could not unregister as started bundle" ) );
         }
      }

      // If a bundle activator is specified call its stop method
      // If an error occurs, continue this method and throw a bundle exception
      // after performing the rest of this method
      Throwable exceptionWhileStopping = null;
      try {
         if ( bundleActivator != null ) {
            bundleActivator.stop( bundleContext );
         }
      }
      catch( Throwable e ) {
         exceptionWhileStopping = e;
      }

      // Unregister service/bundle/framework listeners,
      // unregister services registered
      // and release services used
      cleanUp();

      // Set the state to RESOLVED
      state = RESOLVED;

      // Broadcast  bundle stopped event
      getCube().fireBundleEvent( BundleEvent.STOPPED, this );

      // Disable bundle context (so that anyone who has kept a reference on it will
      // get an IllegalStateException when calling methods)
      bundleContext.setStopped();

      // Throw a bundle exception if an error occured in the stop method of the bundle
      // activator
      if ( exceptionWhileStopping != null ) {
         throw new BundleException( "Exception occured in the bundle activator's stop method", exceptionWhileStopping );
      }
   }

// *****************************************************************************

   private boolean waitUntilStartedOrStopped() {
      int reasonableTime = getCube().getReasonableTimeToWait();
      long stopTime = System.currentTimeMillis() + reasonableTime;
      int sleepTime = reasonableTime / 10;
      while ( ( getState() == STARTING || getState() == STOPPING ) && System.currentTimeMillis() < stopTime ) {
         try {
            Thread.sleep( sleepTime );
         }
         catch ( InterruptedException ignore ) {}
      }
      if ( getState() != STARTING && getState() != STOPPING ) {
         return true;
      }
      else {
         return false;
      }
   }

   // Called when the bundle is created or updated
   private void initializeBundle( String location, File bundleJarFile ) throws IOException, ClassNotFoundException {
      this.location = location;
      this.bundleJarFile = bundleJarFile;

      // Initialize headers
      ZipFile archive = new ZipFile( bundleJarFile );
      ManifestHelper manifest = new ManifestHelper( archive );
      headers = new BundleHeaders( manifest.getMainAttributes() );
      archive.close();

      // Cannot create instance of bundle activator or even load bundle activator class
      // here because that might require classes that are not available yet if the bundle
      // is not resolved
      // That is done in in the start method, when we are sure the bundle is resolved
      bundleActivator = null;

      // Set state to "installed"
      state = INSTALLED;

      // Resolve classpath and native code dependencies if any
      NamedPropertySet[] classPath = NamedPropertySet.createNamedPropertySets( (String)headers.get( Constants.BUNDLE_CLASSPATH ) );
      NamedPropertySet[] nativeLibraries = NamedPropertySet.createNamedPropertySets( (String)headers.get( Constants.BUNDLE_NATIVECODE ) );
      classLoader = getCube().createClassLoaderForBundle( this, bundleJarFile, nativeLibraries, classPath );

      // Register package imports and exports if any
      exportedPackages = NamedPropertySet.createNamedPropertySets( (String)headers.get( Constants.EXPORT_PACKAGE ) );
      if ( exportedPackages != null ) {
         getCube().checkPackagePermission( exportedPackages, PackagePermission.EXPORT, classLoader.getCodeSource() );
      }
      else {
         exportedPackages = new NamedPropertySet[ 0 ];
      }
      explicitlyImportedPackages = NamedPropertySet.createNamedPropertySets( (String)headers.get( Constants.IMPORT_PACKAGE ) );
      if ( explicitlyImportedPackages != null ) {
         getCube().checkPackagePermission( explicitlyImportedPackages, PackagePermission.IMPORT, classLoader.getCodeSource() );
      }
      else {
         explicitlyImportedPackages = new NamedPropertySet[ 0 ];
      }

      // exported packages are the implicitly imported packages
      // and the explicitly exported packages
      //
      // the implicitly imported packages are the exported packages,
      // except when an exported package is imported in a different version than exported
      // (it can make sense to import a lower version that exported, I don't know if it
      // makes sense to import a higher version than exported)
      Vector vec = new Vector();
      for ( int i = 0; i < explicitlyImportedPackages.length; i++ ) {
         vec.addElement( explicitlyImportedPackages[ i ] );
      }
      for ( int i = 0; i < exportedPackages.length; i++ ) {
         boolean explicitlyExported = false;
         for ( int j = 0; j < explicitlyImportedPackages.length; j++ ) {
            if ( exportedPackages[ i ].getName().equals( explicitlyImportedPackages[ j ].getName() ) ) {
               explicitlyExported = true;
            }
         }
         if ( !explicitlyExported ) {
            vec.addElement( exportedPackages[ i ] );
         }
      }
      NamedPropertySet[] importedPackages = new NamedPropertySet[ vec.size() ];
      vec.copyInto( importedPackages );

      for ( int i = 0; i < exportedPackages.length; i++ ) {
         getCube().getPackageAdmin().exportPackage( this, exportedPackages[ i ].getName(), (String)exportedPackages[ i ].getProperty( Constants.PACKAGE_SPECIFICATION_VERSION ), classLoader );
      }
      for ( int i = 0; i < importedPackages.length; i++ ) {
         getCube().getPackageAdmin().importPackage( this, importedPackages[ i ].getName(), (String)importedPackages[ i ].getProperty( Constants.PACKAGE_SPECIFICATION_VERSION ) );
      }
   }

   // Called when the bundle is uninstalled or in first phase of bundle update
   private void uninitializeBundle() throws IOException {
      // no need to call cleanUp() because stop() must have been called before we get here.

      bundleJarFile = null;
      bundleActivator = null;

      getCube().getPackageAdmin().bundleUpdatedOrUninstalled( this );
   }

   // Called when the bundle stops
   private void cleanUp() {
      // remove bundle listeners
      getCube().removeServiceListenersRegisteredBy( this );
      getCube().removeBundleListenersRegisteredBy( this );
      getCube().removeFrameworkListenersRegisteredBy( this );

      // unregister services registered
      ServiceRegistrationImpl[] serviceRegistrations = getCube().getServiceRegistry().getServicesRegisteredBy( this );
      if ( serviceRegistrations != null ) {
         for ( int i = 0; i < serviceRegistrations.length; i++ ) {
            serviceRegistrations[ i ].unregister();
         }
      }

      // release services used
// could call getCube().getServiceRegistry().ungetAllServices( bundle )
      ServiceReferenceImpl[] serviceReferences = getCube().getServiceRegistry().getServicesUsedBy( this );
      if ( serviceReferences != null ) {
         for ( int i = 0; i < serviceReferences.length; i++ ) {
            getCube().getServiceRegistry().ungetService( serviceReferences[ i ], this );
         }
      }
   }
/*
   protected void eagerlyRemove() throws IOException {
      // unregister exported packages
      getCube().getPackageRegistry().unregisterExportedPackages( this );
      // see if any bundles need to be unresolved since we removed exported packages
      getCube().unresolveBundles();

      // clean up class loader
      classLoader.cleanUp();
      classLoader = null;

      // clean up old class loaders
      if ( oldClassLoaders != null ) {
         for ( int i = 1; i < oldClassLoaders.size(); i++ ) {
            ( (BundleClassLoader)oldClassLoaders.elementAt( i ) ).cleanUp();
         }
         oldClassLoaders = null;
      }
   }
*/
}