/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube;

import org.osgi.framework.Bundle;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

// A utility class used to keep a mapping between bundles that register event listeners
// and the event listeners that are registered.
// Bundles are responsible for registering and unregistering their listeners.
public abstract class ListenerSet {

   // Hashtable of listeners:   registering bundle --> vector of listeners
   private Hashtable listeners;

   public ListenerSet() {
      // Create the hashtable of listeners.
      listeners = new Hashtable();
   }

   // Add a listener on behalf of a bundle.
   protected synchronized void addListener( Bundle bundle, Object listener ) {
      Vector v = getListenersRegisteredBy( bundle, true );
      // Do not register listener multiple times
      if ( v.indexOf( listener ) == -1 ) {
         v.addElement( listener );
      }
   }

   // Remove a listener of on behalf of a bundle.
   protected synchronized void removeListener( Bundle bundle, Object listener ) {
      getListenersRegisteredBy( bundle, true ).removeElement( listener );
   }

   // Remove all listeners registered by the specified bundle.
   protected synchronized void removeAllListenersRegisteredBy( Bundle bundle ) {
      listeners.remove( bundle );
   }

   // Return a list of all the listeners registered registered.
   protected synchronized Enumeration getAllListeners() {
      Enumeration enum = listeners.elements();
      Vector v;
      // Create a vector to hold the list of listeners.
      Vector tmp = new Vector();
      // For each event source, add each of its listeners to the list of listeners.
      while ( enum.hasMoreElements() ) {
         v = (Vector)enum.nextElement();
         for ( int i = 0; i < v.size(); i++ ) {
            tmp.addElement( v.elementAt( i ) );
         }
      }
      return tmp.elements();
   }

   // Return a list of all the bundles that have registered listeners.
   protected synchronized Enumeration getBundles() {
      return ( (Hashtable)listeners.clone() ).keys();
   }

   // Return a list of all the listeners registered by the specified bundle.
   // If no listener has been registered by the specified bundle, return an empty
   // list.
   protected synchronized Object[] getListenersRegisteredBy( Bundle bundle ) {
      // Get the list of listeners registered by the given bundle.
      Vector v = getListenersRegisteredBy( bundle, false );
      // If the list of listeners is null, create it.
      if ( v == null ) {
         v = new Vector();
      }
      // Return the list elements in the form of an array.
      Object[] tmp = new Object[ v.size() ];
      v.copyInto( tmp );
      return tmp;
   }

   // Return a list of all the listeners registered by a specified bundle.
   //
   // If no listener has been registered by the the specified bundle, then:
   //    if createIfNull is false, return null
   //    if createIfNull is true, create a new list of listeners for the bundle and return
   //       that empty list
   //
   // This method is for the private use of this class; the listeners registered by a given
   // bundle can be obtained by other classes by calling the other getListenersRegisteredBy
   // method, which takes a single argument (the bundle).
   protected synchronized Vector getListenersRegisteredBy( Bundle bundle, boolean createIfNull ) {
      // If no listener list exists for the bundle and if createIfNull is true, create
      // a new list of listeners for the given bundle.
      if ( listeners.get( bundle ) == null && createIfNull ) {
         listeners.put( bundle, new Vector() );
      }
      // Return the list of listeners registered by the bundle, or null if one does not exist.
      return (Vector)listeners.get( bundle );
   }

}