/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;

// This class represents a named property set, i.e. a set of properties and a name.
// It is used to hold parts of a bundle manifest header entry.
// A bundle manifest header entry takes the form:
//    headerTitle: name1; propName11=propValue11; propName12=propValue12,
//                 name2; propName21=propValue21,
//                 name3, ...
// Thus a bundle manifest header entry is made up of a title and one or more value parts
// each of which contains a name and possibly a set of properties.
// This class is used to hold the data of one of the parts of the entry, i.e. one name and
// its set of properties.
// Thus a bundle manifest header entry can be represented by a title and one or more instances
// of NamedPropertySet.
public class NamedPropertySet {

   // A special property that is found is bundle manifest header entries is the specification
   // version of exported and imported packages.
   public static final String SPECIFICATION_VERSION = "specification-version";

   // the name of the property set
   private String name;
   // the set of properties
   private Hashtable propertySet;

   // Create a list of named property sets from a string.  The string that is passed
   // here could be the part of a bundle manifest header entry after the header name
   // and colon.
   // The string must have the form that is used for bundle manifest header entries:
   //   SET (, SET)*
   //   where  SET = name (;PARAM)*
   //          PARAM = paramName '=' paramValue
   public static NamedPropertySet[] createNamedPropertySets( String description ) {
      // If the string is null, return null.
      if ( description == null ) {
         return null;
      }

      // Decompose the bundle manifest entry string into its parts
      StringTokenizer tokens = new StringTokenizer( description, "," );
      Vector tmp = new Vector();
      String token;
      while ( tokens.hasMoreTokens() ) {
         token = tokens.nextToken();
         // For each part, create a NamedPropertySet.
         tmp.addElement( createNamedPropertySet( token ) );
      }
      // Create the array of named property sets and return it.
      NamedPropertySet[] ret = new NamedPropertySet[ tmp.size() ];
      tmp.copyInto( ret );
      return ret;
   }

   // Create a named property set from a string.  The string that is passed
   // here could be a sub-part of a bundle manifest header entry.
   // The string must have the form that is used for bundle manifest header entry parts:
   //   name (; PARAM)*
   //   where  PARAM = paramName '=' paramValue
   public static NamedPropertySet createNamedPropertySet( String description ) {
      // If the string is null, return null.
      if ( description == null ) {
         return null;
      }

      String name;
      Hashtable propertySet = new Hashtable();

      // Decompose the description into its parts.
      StringTokenizer tokens = new StringTokenizer( description, ";" );
      // The name is the first item found in the string.
      name = tokens.nextToken().trim();
      // The name should not contain an equals sign; otherwise it is a property.
      int n = name.indexOf( "=" );
      if ( n != -1 ) {
         throw new IllegalArgumentException( "= sign found in name" );
      }

      String token;
      // All the parts after the name are properties.
      // We find the equals sign, and then determine the name and the value of the
      // property.
      while ( tokens.hasMoreTokens() ) {
         token = tokens.nextToken().trim();
         n = token.indexOf( "=" );
         if ( n < 0 ) {
            throw new IllegalArgumentException( "= expected but not found" );
         }
         if ( n < 1 ) {
            throw new IllegalArgumentException( "No property name found" );
         }
         if ( n == token.length() ) {
            throw new IllegalArgumentException( "No property value found" );
         }
         propertySet.put( token.substring( 0, n ), token.substring( n + 1, token.length() ) );
      }

      // Create the named property set and return it.
      return new NamedPropertySet( name, propertySet );
   }

   public NamedPropertySet( String name, Hashtable propertySet ) {
      this.name = name;
      this.propertySet = propertySet;
   }

   public String getName() {
      return name;
   }

   public Enumeration getPropertyNames() {
      return propertySet.keys();
   }

   public Object getProperty( String key ) {
      return propertySet.get( key );
   }

}
