/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube;

import org.osgi.framework.ServiceReference;
import org.osgi.framework.Bundle;
import org.osgi.framework.Constants;

import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

// Implementation of the org.osgi.framework.ServiceReference interface
public class ServiceReferenceImpl implements ServiceReference {

   private static long lastAssignedId = -1;

   // the bundle that registered this service
   private Bundle bundle;
   // the class names under which this service was registered
   private String[] objectClass;
   // the properties this service was registered with
   private long id;
   private Hashtable properties;
   // the list of bundle currently using this service
   private Vector usingBundles;

   public ServiceReferenceImpl( Bundle bundle, String[] objectClass, Dictionary props ) {
      this.bundle = bundle;
      this.objectClass = objectClass;

      id = lastAssignedId + 1;
      lastAssignedId = id;

      setProperties( props );

      usingBundles = new Vector();
   }

   // Return a property of this service.
   // Keep returning properties even after the service is unregistered.
   public Object getProperty( String key ) {
      Util.checkNullParameter( key, getClass().getName(), "getProperty", "key" );

      String k = getPropertyKey( key );
      if ( k != null ) {
         return properties.get( k );
      }
      return null;
   }

   // Return the property keys for this service.
   // Keep returning the property keys even after the service is unregistered.
   public String[] getPropertyKeys() {
      String[] keys = new String[ properties.size() ];
      Enumeration enum = properties.keys();
      int i = 0;
      while ( enum.hasMoreElements() ) {
         keys[ i++ ] = (String)enum.nextElement();
      }
      return keys;
   }

   // Return the service that registered this bundle while the service is registered.
   // Return null if the service has been unregistered.
   public Bundle getBundle() {
      // This class's bundle attribute is set by the constructor.
      // When the service is unregistered, the service registration calls this class's
      // serviceUnregistered() method, which sets the bundle attribute to null.
      return bundle;
   }

   public Bundle[] getUsingBundles() {
      Vector tmp = (Vector)usingBundles.clone();
      Bundle[] ret = new Bundle[ tmp.size() ];
      tmp.copyInto( ret );
      return ret;
   }

// *****************************************************************************

   // This method is called by the service registration's set property method.
   protected void setProperties( Dictionary props ) {
      // Create a new hashtable of properties.
      Hashtable newProperties = new Hashtable();
      // Copy the values of the props argument (if any) into the hashtable
      if ( props != null ) {
         Enumeration propNames = props.keys();
         String propName;
         Object propValue;
         while ( propNames.hasMoreElements() ) {
            propName = (String)propNames.nextElement();
            propValue = props.get( propName );
            if ( propName.equalsIgnoreCase( Constants.SERVICE_ID ) ) {
               // osgi_compliance
               // should not normally allow setting service id
               // but framework.property.tc1 osgi compliance test fails if we throw an
               // exception here
               //throw new IllegalArgumentException( propName + " is a reserved property" );
            }
            if ( propName.equalsIgnoreCase( Constants.SERVICE_RANKING ) ) {
               if ( !( propValue instanceof Integer ) ) {
                  if ( propValue instanceof String ) {
                     try {
                        Integer.valueOf( (String)propValue ).intValue();
                     }
                     catch ( NumberFormatException e ) {
                        throw new IllegalArgumentException( "Value of property named '" + propName + "' must be an integer" );
                     }
                  }
                  else {
                     throw new IllegalArgumentException( "Value of property named '" + propName + "' must be an integer" );
                  }
               }
            }
            // if property exists with same key but different case, throw exception
            String caseVariant = getPropertyKey( propName );
            if ( caseVariant != null && !caseVariant.equals( propName ) ) {
               throw new IllegalArgumentException( "Property with case-variant of the same key name already exists" );
            }
            newProperties.put( propName, propValue );
         }
      }
      // Make the "objectClass" property equal to the array of class names
      // under which this service was registered
      newProperties.put( Constants.OBJECTCLASS, objectClass );
      newProperties.put( Constants.SERVICE_ID, new Long( id ) );

      // set new properties only if method ended without exceptions
      properties = newProperties;
   }

   // This method is called by the service registration when the service is unregistered.
   // The service reference's getBundle() method must return null after the service has been
   // unregistered.
   protected void serviceUnregistered() {
      bundle = null;
   }

   // Utility method to get the "objectClass" property without going through the properties.
   // The value that is returned by this method is the same as that returned by the
   // getProperty() method with a property name equal to "objectClass".
   protected String[] getObjectClass() {
      return objectClass;
   }

   // This method should be called ONLY by ServiceRegistryEntry.addUser()
   // If the using bundle uses this service multiple times simultaneously,
   // this method is called only once when the service use count for the bundle
   // reaches one
   public void addUsingBundle( Bundle bundle ) {
      // this check should not be necessary, just to make extra sure
      // we don't add the bundle multiple times
      if ( usingBundles.indexOf( bundle ) == -1 ) {
         usingBundles.addElement( bundle );
      }
   }

   // This method should be called ONLY by ServiceRegistryEntry.removeUser()
   // If the using bundle uses this service multiple times simultaneously,
   // this method is called only once when the service use count for the bundle
   // reaches zero
   public void removeUsingBundle( Bundle bundle ) {
      usingBundles.removeElement( bundle );
   }

   // Return the property key in the case variant that it was registered
   // (will be null if no such property key has been registered, or a string that
   // is a case-variant of key argument).
   private String getPropertyKey( String key ) {
      // properties can be null the first time setProperties() is called
      if ( properties == null ) {
         return null;
      }

      Enumeration keys = properties.keys();
      String k;
      while ( keys.hasMoreElements() ) {
         k = (String)keys.nextElement();
         if ( k.equalsIgnoreCase( key ) ) {
            return k;
         }
      }
      return null;
   }

}