/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 2002, 2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */
 
package com.opensugar.cube;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

// Utility class that contains utility methods used by other classes
public abstract class Util {

   // Check if a parameter value is null, and throw an IllegalArgumentException if it is.
   protected static void checkNullParameter( Object parameter, String className, String methodName, String parameterName ) {
      if ( parameter == null ) {
         throw new IllegalArgumentException( "Cannot call " + className + "." + methodName + " with null " + parameterName );
      }
   }

   // Check if an array parameter value is null, and throw an IllegalArgumentException if it is.
   // Also check the elements of the array, and throw an IllegalArgumentException if one of them is.
   protected static void checkNullArrayParameter( Object[] parameter, String className, String methodName, String parameterName ) {
      if ( parameter == null ) {
         throw new IllegalArgumentException( "Cannot call " + className + "." + methodName + " with null " + parameterName );
      }
      for ( int i = 0; i < parameter.length; i++ ) {
         if ( parameter[ i ] == null ) {
            throw new IllegalArgumentException( "Cannot call " + className + "." + methodName + " with a " + parameterName + " parameter that contains null values" );
         }
      }
   }

   // Delete a file or directory.
   // If the argument is a directory, delete the whole file tree below the directory.
   // Return true if deletion successful, false if not.
   public static boolean recursiveFileDelete( File file ) {
      if ( !file.exists() ) {
         return true;
      }

      if ( file.isDirectory() ) {
         String[] listing = file.list();
         for ( int i = 0; i < listing.length; i++ ) {
            if ( !recursiveFileDelete( new File( file, listing[ i ] ) ) ) {
               return false;
            }
         }
      }
      return file.delete();
   }

   // Recursively copy a directory
   public static boolean recursiveFileCopy( File source, File target ) {
      if ( !source.exists() ) {
         return false;
      }

      if ( source.isDirectory() ) {
         if ( !target.exists() ) {
            target.mkdirs();
         }
         String[] listing = source.list();
         for ( int i = 0; i < listing.length; i++ ) {
            if ( !recursiveFileCopy( new File( source, listing[ i ] ), new File( target, listing[ i ] ) ) ) {
               return false;
            }
         }
      }
      else {
         try {
            transferData( new FileInputStream( source ), new FileOutputStream( target ) );
         }
         catch ( IOException e ) {
            return false;
         }
      }

      return true;
   }

   // Copy data from an input stream into an output stream.
   // Close both streams when done, even if an exception occurs during data transfer.
   public static void transferData( InputStream source, OutputStream destination ) throws IOException {
      BufferedInputStream bis = new BufferedInputStream( source );
      BufferedOutputStream bos = new BufferedOutputStream( destination );
      byte[] buffer = new byte[ 4096 ];
      int n;
      IOException exception = null;
      try {
         while ( ( n = bis.read( buffer ) ) != -1 ) {
            bos.write( buffer, 0, n );
         }
      }
      catch( IOException e ) {
         exception = e;
      }
      finally {
         bis.close();
         bos.close();
      }

      if ( exception != null ) {
         throw exception;
      }
   }

   public static Number[] sortNumbers( Number[] unsorted ) {
      Vector tmp = new Vector();
      for ( int i = 0; i < unsorted.length; i++ ) {
         tmp.addElement( unsorted[ i ] );
      }
      tmp = sortNumbers( tmp );
      Number[] sorted = new Number[ unsorted.length ];
      tmp.copyInto( sorted );
      return sorted;
   }

   private static Vector sortNumbers( Vector v ) {
      Vector sorted = new Vector();
      int n;
      while ( v.size() > 0 ) {
         n = findIndexOfMinNumber( v );
         sorted.addElement( v.elementAt( n ) );
         v.removeElementAt( n );
      }
      return sorted;
   }

   private static int findIndexOfMinNumber( Vector v ) {
      int index = 0;
      Number minVal;
      Number val;
      minVal = (Number)v.elementAt( 0 );
      for ( int i = 1; i < v.size(); i++ ) {
         val = (Number)v.elementAt( i );
         if ( val.doubleValue() < minVal.doubleValue() ) {
            index = i;
            minVal = val;
         }
      }
      return index;
   }

}