/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube.java2;

import java.io.FilterInputStream;
import java.io.DataOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.zip.ZipFile;
import java.util.Properties;

public class ManifestHelper3 {

   private static final String MANIFEST_FILE = "META-INF/MANIFEST.MF";

   // manifest main attributes
   private Properties attributes = new Properties();

   /**
    * Constructs a new Manifest from the specified input stream.
    *
    * @param is the input stream containing manifest data
    * @throws IOException if an I/O error has occured
    */
   public ManifestHelper3( ZipFile archive ) throws IOException {
      InputStream is = archive.getInputStream( archive.getEntry( MANIFEST_FILE ) );
      read( is );
   }

   /**
    * Returns the main Attributes for the Manifest.
    * @return the main Attributes for the Manifest
    */
   public Properties getMainAttributes() {
      return attributes;
   }

   /**
    * Reads the Manifest from the specified InputStream. The entry
    * names and attributes read will be merged in with the current
    * manifest entries.
    *
    * @param is the input stream
    * @exception IOException if an I/O error has occurred
    */
   public void read( InputStream is ) throws IOException {
      // Buffered input stream for reading manifest data
	   FastInputStream fis = new FastInputStream( is );
	   // Line buffer
	   byte[] lbuf = new byte[ 512 ];
	   // Read the main attributes for the manifest
	   readAttributes( fis, lbuf );

      // Do not treat entries.
   }

   private void readAttributes( FastInputStream is, byte[] lbuf ) throws IOException {
   	String name = null, value = null;
	   int len;
      while ( ( len = is.readLine( lbuf ) ) != -1 ) {
         if ( lbuf[ --len ] != '\n' ) {
		      throw new IOException( "line too long" );
	      }
         if ( len > 0 && lbuf[ len-1 ] == '\r' ) {
            --len;
         }
	      if ( len == 0 ) {
            break;
	      }
	      int i = 0;
	      if ( lbuf[0] == ' ' ) {
            // continuation of previous line
            if ( name == null ) {
		         throw new IOException( "misplaced continuation line" );
            }
		      value = value + new String( lbuf, 0, 1, len - 1 );
         }
         else {
            while ( lbuf[ i++ ] != ':' ) {
               if ( i >= len ) {
                  throw new IOException("invalid header field");
               }
            }
            if ( lbuf[ i++ ] != ' ' ) {
               throw new IOException("invalid header field");
            }
            name = new String(lbuf, 0, 0, i - 2);
            value = new String(lbuf, 0, i, len - i);
         }

         try {
System.out.println( "manifest attr: " + name + ": _" + value + "_" );
            attributes.put( name, value );
         }
         catch ( IllegalArgumentException e ) {
            throw new IOException( "invalid header field name: " + name );
         }
      }
   }


   static class FastInputStream extends FilterInputStream {

      private byte buf[];
   	private int count = 0;
	   private int pos = 0;

   	FastInputStream( InputStream in ) {
         this( in, 8192 );
   	}

	   FastInputStream( InputStream in, int size ) {
         super( in );
         buf = new byte[ size ];
	   }

   	public int read() throws IOException {
         if ( pos >= count ) {
   	   	fill();
   		   if ( pos >= count ) {
               return -1;
	   	   }
         }
         return buf[ pos++ ] & 0xff;
	   }

   	public int read( byte[] b, int off, int len ) throws IOException {
         int avail = count - pos;
         if ( avail <= 0 ) {
		      if ( len >= buf.length ) {
               return in.read( b, off, len );
      		}
	      	fill();
	   	   avail = count - pos;
      		if ( avail <= 0 ) {
               return -1;
		      }
         }
         if ( len > avail ) {
		      len = avail;
         }
         System.arraycopy( buf, pos, b, off, len );
         pos += len;
         return len;
   	}

   	/*
	    * Reads 'len' bytes from the input stream, or until an end-of-line
   	 * is reached. Returns the number of bytes read.
	    */
   	public int readLine( byte[] b, int off, int len ) throws IOException {
         byte[] tbuf = this.buf;
         int total = 0;
         while ( total < len ) {
		      int avail = count - pos;
	   	   if ( avail <= 0 ) {
               fill();
               avail = count - pos;
               if ( avail <= 0 ) {
                  return -1;
               }
   		   }
		      int n = len - total;
		      if ( n > avail ) {
               n = avail;
   		   }
		      int tpos = pos;
		      int maxpos = tpos + n;
	   	   while ( tpos < maxpos && tbuf[ tpos++ ] != '\n' ) ;
            n = tpos - pos;
            System.arraycopy( tbuf, pos, b, off, n );
		      off += n;
	   	   total += n;
   		   pos = tpos;
		      if ( tbuf[ tpos-1 ] == '\n' ) {
               break;
            }
         }
         return total;
	   }

   	public byte peek() throws IOException {
         if ( pos == count ) {
	   	   fill();
         }
         return buf[ pos ];
	   }

   	public int readLine( byte[] b ) throws IOException {
         return readLine( b, 0, b.length );
   	}

	   public long skip( long n ) throws IOException {
         if ( n <= 0 ) {
		      return 0;
         }
         long avail = count - pos;
         if ( avail <= 0 ) {
		      return in.skip( n );
         }
         if ( n > avail ) {
		      n = avail;
         }
         pos += n;
         return n;
	   }

   	public int available() throws IOException {
         return ( count - pos ) + in.available();
	   }

   	public void close() throws IOException {
         if ( in != null ) {
	   	   in.close();
      		in = null;
	   	   buf = null;
         }
	   }

   	private void fill() throws IOException {
         count = pos = 0;
         int n = in.read( buf, 0, buf.length );
         if ( n > 0 ) {
            count = n;
         }
   	}
   }

}
