/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube.ldap;

import org.osgi.framework.InvalidSyntaxException;

import java.util.Vector;

public abstract class Filter {

   protected static final String AND = "&";
   protected static final String OR = "|";
   protected static final String NOT = "!";

   protected static final String EQUALS = "=";
   protected static final String APPROX = "~=";
   protected static final String GREATER = ">=";
   protected static final String LESS = "<=";

   private static final char ESCAPE_CHAR = '\\';

   public static Clause parse( String filter, boolean keysCaseSensitive ) throws InvalidSyntaxException {
      try {
         return parseFilter( filter, keysCaseSensitive, false );
      }
      catch ( InvalidSyntaxException e ) {
         throw new InvalidSyntaxException( e.getMessage(), filter );
      }
   }

   private static Clause parseFilter( String filter, boolean keysCaseSensitive, boolean not ) throws InvalidSyntaxException {
      if ( filter == null || filter.trim().length() == 0 ) {
         if ( !not ) {
            return new NullClause();
         }
         else {
            throw new InvalidSyntaxException( "Not operator used on an empty filter", filter );
         }
      }

      filter = filter.trim();
      if ( filter.startsWith( "(" ) && filter.endsWith( ")" ) ) {
         Clause clause = parseFilterComp( filter.substring( 1, filter.length() - 1 ), keysCaseSensitive );
         if ( not ) {
            clause = new NotClause( clause );
         }
         return clause;
      }
      throw new InvalidSyntaxException( "Missing opening or closing bracket on filter", filter );
   }

   private static Clause parseFilterComp( String filterComp, boolean keysCaseSensitive ) throws InvalidSyntaxException {
      // remove leading spaces - but not trailing spaces which are useful characters
      // if filterComp is an item
      while ( Character.isWhitespace( filterComp.charAt( 0 ) ) ) {
         filterComp = filterComp.substring( 1 );
      }

      if ( filterComp.startsWith( AND ) ) {
         return parseFilterList( filterComp.substring( 1, filterComp.length() ).trim(), AND, keysCaseSensitive );
      }
      else if ( filterComp.startsWith( OR ) ) {
         return parseFilterList( filterComp.substring( 1, filterComp.length() ).trim(), OR, keysCaseSensitive );
      }
      else if ( filterComp.startsWith( NOT ) ) {
         return parseFilter( filterComp.substring( 1, filterComp.length() ).trim(), keysCaseSensitive, true );
      }
      else {
         return parseItem( filterComp, keysCaseSensitive );
      }
   }

   private static Clause parseFilterList( String filterList, String op, boolean keysCaseSensitive ) throws InvalidSyntaxException {
      Clause[] clauses = parseFilterList( filterList, keysCaseSensitive );

      if ( op.equals( AND ) ) {
         return new SuperClauseAND( clauses );
      }
      else if ( op.equals( OR ) ) {
         return new SuperClauseOR( clauses );
      }
      else {
         throw new InvalidSyntaxException( "Invalid boolean operator on filter list", null );
      }
   }

   private static Clause parseItem( String s, boolean keysCaseSensitive ) throws InvalidSyntaxException {
      int n1 = s.indexOf( EQUALS );
      int n2 = s.indexOf( APPROX );
      int n3 = s.indexOf( GREATER );
      int n4 = s.indexOf( LESS );
      if ( n1 == -1 ) {
         n1 = Integer.MAX_VALUE;
      }
      if ( n2 == -1 ) {
         n2 = Integer.MAX_VALUE;
      }
      if ( n3 == -1 ) {
         n3 = Integer.MAX_VALUE;
      }
      if ( n4 == -1 ) {
         n4 = Integer.MAX_VALUE;
      }

      String filterType;
      int n, m;
      if ( n1 < n2 && n1 < n3 && n1 < n4 ) {
         filterType = EQUALS;
         n = n1;
         m = n + 1;
      }
      else if ( n2 < n3 && n2 < n4 ) {
         filterType = APPROX;
         n = n2;
         m = n + 2;
      }
      else if ( n3 < n4 ) {
         filterType = GREATER;
         n = n3;
         m = n + 2;
      }
      else {
         filterType = LESS;
         n = n4;
         m = n + 2;
      }

      if ( n > s.length() || m > s.length() ) {
         throw new InvalidSyntaxException( "No filter type found in filter comp", null );
      }

      return createClause( filterType, s.substring( 0, n ), s.substring( m, s.length() ), keysCaseSensitive );
   }

   private static Clause[] parseFilterList( String filterList, boolean keysCaseSensitive ) throws InvalidSyntaxException {
      filterList = filterList.trim();

      if ( filterList.length() == 0 ) {
         throw new InvalidSyntaxException( "Empty filter list", null );
      }
      Vector v = new Vector();
      String filter;
      while ( filterList.length() > 0 ) {
         filter = extractFirstFilter( filterList );
         v.addElement( parseFilter( filter, keysCaseSensitive, false ) );
         filterList = filterList.substring( filter.length(), filterList.length() ).trim();
      }

      if ( v.size() < 2 ) {
         throw new InvalidSyntaxException( "Filter list with a single filter", null );
      }

      Clause[] clauses = new Clause[ v.size() ];
      for ( int i = 0; i < v.size(); i++ ) {
         clauses[ i ] = (Clause)v.elementAt( i );
      }
      return clauses;
   }

   private static String extractFirstFilter( String filterList ) throws InvalidSyntaxException {
      filterList = filterList.trim();

      if ( filterList.startsWith( "(" ) ) {
         int n = filterList.indexOf( ')' );
         if ( n == -1 ) {
            throw new InvalidSyntaxException( "Missing closing bracket", null );
         }
         return filterList.substring( 0, n + 1 );
      }
      throw new InvalidSyntaxException( "Missing opening bracket", null );
   }

   private static Clause createClause( String filterType, String attr, String value, boolean keysCaseSensitive ) throws InvalidSyntaxException {
      attr = attr.trim();

      // There should be no unescaped '(', ')', '<', '>' '~' or '=' characters in the attr
      if ( getIndicesOfNonEscapedChar( attr, '(' ).length > 0 ||
           getIndicesOfNonEscapedChar( attr, ')' ).length > 0 ||
           getIndicesOfNonEscapedChar( attr, '<' ).length > 0 ||
           getIndicesOfNonEscapedChar( attr, '>' ).length > 0 ||
           getIndicesOfNonEscapedChar( attr, '~' ).length > 0 ||
           getIndicesOfNonEscapedChar( attr, '=' ).length > 0 ) {
         throw new InvalidSyntaxException( "Special character not escaped (in attr)", null );
      }
      // There should be no unescaped '(' or ')' characters in the value
      if ( getIndicesOfNonEscapedChar( value, '(' ).length > 0 ||
           getIndicesOfNonEscapedChar( value, ')' ).length > 0 ) {
         throw new InvalidSyntaxException( "Special character not escaped (in value)", null );
      }
      // There should be no unescaped WILDCARD characters in the value
      // unless filter is of type EQUALS
      if ( !filterType.equals( EQUALS ) &&
           getIndicesOfNonEscapedChar( value, SubstringClause.WILDCARD ).length > 0 ) {
         throw new InvalidSyntaxException( "Special character not escaped (in value with equal filter)", null );
      }

      String valueNoEscapeChars = removeEscapeChars( value );
      try {
         if ( filterType.equals( EQUALS ) ) {
            if ( value.equals( "" + SubstringClause.WILDCARD ) ) {
               return new PresenceClause( attr, keysCaseSensitive );
            }
            else if ( getIndicesOfNonEscapedChar( value, SubstringClause.WILDCARD ).length > 0 ) {
               return new SubstringClause( attr, valueNoEscapeChars, keysCaseSensitive );
            }
            return new EqualsClause( attr, valueNoEscapeChars, keysCaseSensitive );
         }
         else if ( filterType.equals( APPROX ) ) {
            return new ApproxClause( attr, valueNoEscapeChars, keysCaseSensitive );
         }
         else if ( filterType.equals( GREATER ) ) {
            return new GreaterThanClause( attr, valueNoEscapeChars, keysCaseSensitive );
         }
         else if ( filterType.equals( LESS ) ) {
            return new LessThanClause( attr, valueNoEscapeChars, keysCaseSensitive );
         }
         else {
            throw new InvalidSyntaxException( "Invalid filter type", null );
         }
      }
      catch( IllegalArgumentException e ) {
         throw new InvalidSyntaxException( e.getMessage(), null );
      }
   }

   private static int[] getIndicesOfNonEscapedChar( String s, char c ) {
      int n = -1;
      Vector v = new Vector();
      while ( ( n = s.indexOf( c, n + 1 ) ) != -1 ) {
         if ( n == 0 || s.charAt( n - 1 ) != ESCAPE_CHAR ) {
            v.addElement( new Integer( n ) );
         }
      }
      int[] ret = new int[ v.size() ];
      for ( int i = 0; i < v.size(); i++ ) {
         ret[ i ] = ( (Integer)v.elementAt( i ) ).intValue();
      }
      return ret;
   }

   private static String removeEscapeChars( String s ) {
      s = removeEscapeChar( s, SubstringClause.WILDCARD );
      s = removeEscapeChar( s, '(' );
      return removeEscapeChar( s, ')' );
   }

   private static String removeEscapeChar( String s, char escapedChar ) {
      StringBuffer sb = new StringBuffer();
      int n = 0;
      int m;
      while ( ( m = s.indexOf( "" + ESCAPE_CHAR + escapedChar, n ) ) != -1 ) {
         sb.append( s.substring( n, m ) + escapedChar );
         n = m + 2;
      }
      if ( n < s.length() ) {
         sb.append( s.substring( n, s.length() ) );
      }
      return sb.toString();
   }

}
