/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube.packageAdmin;

import java.util.StringTokenizer;

// This class represents a package version of type Generation.Revision.Correction
// and formatted as '<int>[.<int>[.<int>]]'.
public class GoRoCoVersion {

   // Undefined version:
   //      generation = null
   //      revision = Integer.MIN_VALUE
   //      correction = Integer.MIN_VALUE

   // If revision or correction not defined, their value is Integer.MIN_VALUE

   // the version generation
   private Integer generation = null;
   // the version revision (if any)
   private Integer revision = new Integer( Integer.MIN_VALUE );
   // the version correction (if any)
   private Integer correction = new Integer( Integer.MIN_VALUE );

   // Value can be null (for an undefined version)
   public GoRoCoVersion( String value ) {
      if ( value != null ) {
         // If version string is wrapped in quotes, remove the quotes.
         if ( ( value.startsWith( "\"" ) && value.endsWith( "\"" ) ) || ( value.startsWith( "'" ) && value.endsWith( "'" ) ) ) {
            value = value.substring( 1, value.length() - 1 );
         }
         // Split the version string into its dot separated parts.
         StringTokenizer tokens = new StringTokenizer( value, "." );
         try {
            // The version string contains at least one part which is the generation.
            generation = Integer.valueOf( tokens.nextToken() );
            // The version string may also contain a revision part.
            if ( tokens.hasMoreTokens() ) {
               revision = Integer.valueOf( tokens.nextToken() );
            }
            // Finally, it may contain one last part which is the correction.
            if ( tokens.hasMoreTokens() ) {
               correction = Integer.valueOf( tokens.nextToken() );
            }
            // If the version string has more than three dot separated parts, it is an invalid
            // version string.
            if ( tokens.hasMoreTokens() ) {
               // ignore further tokens
               //throw new IllegalArgumentException( "Invalid version string: too many tokens" );
            }
         }
         catch( NumberFormatException e ) {
            // If any one of the version string's dot separated parts is not an integer, the version
            // string is ininvalid
            // consider version is null
            generation = null;
            revision = new Integer( Integer.MIN_VALUE );
            correction = new Integer( Integer.MIN_VALUE );
         }
      }
   }

   // Return this version's generation.
   protected Integer getGeneration() {
      return generation;
   }

   // Return this version's revision.
   protected Integer getRevision() {
      return revision;
   }

   // Return this version's correction.
   protected Integer getCorrection() {
      return correction;
   }

   // Implementation of the Comparable interface's compareTo method.
   // Compares this version to another version, returning zero if the two versions are
   // equal, a negative integer if this version is smaller than the version to compare
   // to, and a positive integer otherwise.
   //
   // Note that the value returned gives no indication of how different the versions are;
   // it just gives an indication of which one is larger.
   public int compareTo( Object obj ) {
      GoRoCoVersion versionToCompareTo = (GoRoCoVersion)obj;

      if ( isUndefined() && versionToCompareTo.isUndefined() ) {
         // if both versions are undefined, the versions are considered equal
         return 0;
      }
      else if ( isUndefined() || versionToCompareTo.isUndefined() ) {
         // if either of the versions (but not both) is undefined, this version
         // is considered larger
         return 1;
      }

      Integer generation2 = versionToCompareTo.getGeneration();
      Integer revision2 = versionToCompareTo.getRevision();
      Integer correction2 = versionToCompareTo.getCorrection();

      // Compare generations.
      int n = generation.intValue() - generation2.intValue();
      if ( n == 0 ) {
         // If generations are equal, compare revisions.
         n = revision.intValue() - revision2.intValue();
         if ( n == 0 ) {
            // If generations and revisions are equal, compare corrections.
            return correction.intValue() - correction2.intValue();
         }
         return n;
      }
      return n;
   }

   // Determine if this version is equal to another version.
   public boolean equals( Object obj ) {
      if ( compareTo( obj ) == 0 ) {
         return true;
      }
      return false;
   }

   // Return a string representation of this version.
   public String toString() {
      if ( isUndefined() ) {
         return "Undefined";
      }
      else {
         StringBuffer sb = new StringBuffer( generation.toString() );
         // A negative revision value means a revision is not specified in this version.
         if ( revision.intValue() >= 0 ) {
            sb.append( "." + revision.toString() );
            // A negative correction value means a correction is not specified in this version.
            if ( correction.intValue() >= 0 ) {
               sb.append( "." + correction.toString() );
            }
         }
         return sb.toString();
      }
   }

   // Determine if this version is undefined.
   public boolean isUndefined() {
      if ( generation == null ) {
         return true;
      }
      return false;
   }

}