/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube.packageAdmin;

import com.opensugar.cube.AbstractCube;
import com.opensugar.cube.BundleClassLoader;
import com.opensugar.cube.BundleImpl;
import com.opensugar.cube.NamedPropertySet;

import org.osgi.framework.Bundle;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.Constants;
import org.osgi.service.packageadmin.PackageAdmin;
import org.osgi.service.packageadmin.ExportedPackage;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.net.URL;

public class PackageAdminImpl implements PackageAdmin {

   private AbstractCube cube;

   // exported packages are exported by resolved bundles (therefore really available)
   private Hashtable exportedPackages; // maps package name to ExportedPackageImpl object

   // proposed exports are packages proposed for export by bundles
   private Hashtable proposedExports; // maps bundles to PackageExport objects
   // required imports are package required by bundles
   private Hashtable requiredImports; // maps bundles to PackageImport objects

   public PackageAdminImpl( AbstractCube cube ) {
      this.cube = cube;

      exportedPackages = new Hashtable();
      proposedExports = new Hashtable();
      requiredImports = new Hashtable();
   }

// *****************************************************************************
//
// implements PackageAdmin
//
// *****************************************************************************

   public ExportedPackage getExportedPackage( String packageName ) {
      return (ExportedPackage)exportedPackages.get( packageName );
   }

   public ExportedPackage[] getExportedPackages( Bundle bundle ) {
      Hashtable clone = (Hashtable)exportedPackages.clone();
      Enumeration enum = clone.elements();
      Vector vec = new Vector();
      ExportedPackage exportedPackage;
      while ( enum.hasMoreElements() ) {
         exportedPackage = (ExportedPackage)enum.nextElement();
         if ( bundle == null || bundle.getBundleId() == exportedPackage.getExportingBundle().getBundleId() ) {
            vec.addElement( exportedPackage );
         }
      }
      ExportedPackage[] ret = new ExportedPackage[ vec.size() ];
      vec.copyInto( ret );
      return ret;
   }

   public void refreshPackages( Bundle[] bundles ) {
      final Bundle[] bs = bundles;
      ( new Thread( new Runnable() {
         public void run() {
            doRefreshPackages( bs );
         }
      } ) ).start();
   }


// *****************************************************************************
//
// methods for exporting, importing, unimporting, and refreshing packages
//
// *****************************************************************************

   // Called by a bundle when the bundle is installed or updated,
   // to notify the package registry that the specified package is proposed for export
   // by the bundle
   //
   // Packages are really exported only when the bundle becomes resolved.
   // Resolving bundles is the business of this class's resolveBundles method.
   public synchronized void exportPackage( Bundle exportingBundle, String packageName, String specificationVersion, BundleClassLoader classLoader ) {
      getProposedExports( exportingBundle ).addElement( new PackageExport( packageName, specificationVersion, exportingBundle, classLoader ) );

      // special treatment for system bundle
      // system bundle packages are immediately exported (the system bundle is never
      // in installed state, so it makes no sense for its bundles not to be exported)
      if ( exportingBundle.getBundleId() == 0 ) {
         exportedPackages.put( packageName, new ExportedPackageImpl( packageName, specificationVersion, exportingBundle, classLoader ) );
      }
   }

   // Called by a bundle when the bundle is installed or updated,
   // to notify package registry that the specificed package is required by the bundle
   //
   // Packages are really imported only when the bundle becomes resolved.
   // Resolving bundles is the business of this class's resolveBundles method.
   public synchronized void importPackage( Bundle importingBundle, String packageName, String specificationVersion ) {
      getRequiredImports( importingBundle ).addElement( new PackageImport( importingBundle, packageName, specificationVersion ) );
   }

   // called by a bundle when the bundle is uninstalled or in first stage of bundle update,
   // to notify package registry that the specified bundle no longer participates
   // in bundle sharing
   public synchronized void bundleUpdatedOrUninstalled( BundleImpl bundle ) {
      // remove the specified bundle from the list of importers of all exported packages
      // set the removal pending flag of any package exported by the specified bundle
      Enumeration enum = exportedPackages.elements();
      ExportedPackageImpl exportedPackage;
      while ( enum.hasMoreElements() ) {
         exportedPackage = (ExportedPackageImpl)enum.nextElement();
         if ( bundle.getBundleId() == exportedPackage.getExportingBundle().getBundleId() ) {
            exportedPackage.setRemovalPending();
            // do not remove specified bundle from the list of importing bundles of the
            // exported package (this is an implicit import)
         }
         else {
            // this call will have no effect if the the bundle is not an importer of the package
            exportedPackage.removeImportingBundle( bundle );
         }
      }

      // remove proposed exports and required imports for the bundle
      proposedExports.remove( bundle );
      requiredImports.remove( bundle );
   }


// *****************************************************************************
//
// class and resource loading
//
// *****************************************************************************

   public Class loadClass( Bundle requestingBundle, String fullyQualifiedClassName ) throws ClassNotFoundException {
      // Imported packages cannot be searched by a bundle that is not resolved
      // That is ok, because bundle will not be in the importing bundles of an exported
      // package if it is not resolved

      // Determine the package name from the fully qualified class name.
      int n = fullyQualifiedClassName.lastIndexOf( "." );
      if ( n == -1 ) {
         throw new IllegalArgumentException( "Cannot get a class that does not belong to a package: " + fullyQualifiedClassName );
      }
      String packageName = fullyQualifiedClassName.substring( 0, n );

      // Get the ExportedPackageImpl for the package.
      ExportedPackageImpl exportedPackage = (ExportedPackageImpl)exportedPackages.get( packageName );
      if ( exportedPackage != null ) {
         // Special case for java.security (which need not be explicitly imported but may
         // not be provided by system class loader and may then be exported by the system
         // bundle)
         if ( packageName.equals( "java.security" ) ) {
            exportedPackage.loadClass( fullyQualifiedClassName );
         }
         else if ( exportedPackage.isImportingBundle( requestingBundle ) ) {
            return exportedPackage.loadClass( fullyQualifiedClassName );
         }
      }

      // No exported package found corresponding to class name, or requesting bundle
      // does not import the required package
      throw new ClassNotFoundException( fullyQualifiedClassName );
   }

   public synchronized URL getResource( Bundle requestingBundle, String name ) {
      // Imported packages cannot be searched by a bundle that is not resolved
      // That is ok, because bundle will not be in the importing bundles of an exported
      // package if it is not resolved

      // Determine the package from the resource name.
      int n = name.lastIndexOf( "/" );
      if ( n == -1 ) {
         return null;
      }
      String packageName = name.substring( 0, n );
      packageName = packageName.replace( '/', '.' );
      // Get the ExportedPackageImpl for the package.
      ExportedPackageImpl exportedPackage = (ExportedPackageImpl)exportedPackages.get( packageName );
      if ( exportedPackage != null && exportedPackage.isImportingBundle( requestingBundle ) ) {
         return exportedPackage.getResource( name );
      }

      // No exported package found corresponding to resource name, or requesting bundle
      // does not import the required package
      return null;
   }


// *****************************************************************************
//
// utility methods
//
// *****************************************************************************

   public synchronized PackageImport[] getUnsatisfiedImports( Bundle bundle ) {
      Vector unsatisfied = (Vector)getRequiredImports( bundle ).clone();
      Vector satisfied = new Vector();
      for ( int i = 0; i < unsatisfied.size(); i++ ) {
         PackageImport required = (PackageImport)unsatisfied.elementAt( i );
         ExportedPackageImpl available = (ExportedPackageImpl)exportedPackages.get( required.getName() );
         if ( available != null ) {
            GoRoCoVersion requiredVersion = new GoRoCoVersion( required.getSpecificationVersion() );
            GoRoCoVersion availableVersion = new GoRoCoVersion( available.getSpecificationVersion() );
            // if either available version or required version (but not both) is undefined
            // available version is considered greater
            // see implementation of GoRoCoVersion.compareTo()
            if ( availableVersion.compareTo( requiredVersion ) >= 0 ) {
               satisfied.addElement( required );
            }
         }
      }
      for ( int i = 0; i < satisfied.size(); i++ ) {
         unsatisfied.removeElement( satisfied.elementAt( i ) );
      }

      // Do not include unsatisfied implicit package imports, unless it is for a package
      // that is already exported (in which case if the available version is adequate,
      // the import will already have been removed above, and if the available version
      // is not adequate, then we should include the import)
      Vector implicit = new Vector();
      Vector proposed = getProposedExports( bundle );
      for ( int i = 0; i < unsatisfied.size(); i++ ) {
         PackageImport pack1 = (PackageImport)unsatisfied.elementAt( i );
         if ( exportedPackages.get( pack1.getName() ) == null ) {
            for ( int j = 0; j < proposed.size(); j++ ) {
               PackageExport pack2 = (PackageExport)proposed.elementAt( j );
               if ( pack2 != null && pack2.getName().equals( pack1.getName() ) ) {
                  GoRoCoVersion v1 = new GoRoCoVersion( pack2.getSpecificationVersion() );
                  GoRoCoVersion v2 = new GoRoCoVersion( pack1.getSpecificationVersion() );
                  // if either v1 or required v2 (but not both) is undefined
                  // v1 is considered greater
                  // see implementation of GoRoCoVersion.compareTo()
                  if ( v1.compareTo( v2 ) >= 0 ) {
                     implicit.addElement( pack1 );
                  }
               }
            }
         }
      }
      for ( int i = 0; i < implicit.size(); i++ ) {
         unsatisfied.removeElement( implicit.elementAt( i ) );
      }

      PackageImport[] ret = new PackageImport[ unsatisfied.size() ];
      unsatisfied.copyInto( ret );
      return ret;
   }

// *****************************************************************************
//
// refresh packages
//
// *****************************************************************************

   // called to do package refresh synchronously (used by cube on startup)
   // this is the method that actually does the work
   // the refreshPackages() method just creates a thread in which this method
   // is executed asynchronously
   public synchronized void doRefreshPackages( Bundle[] bundles ) {
      // 1. Compute a graph of bundles starting with the specified ones. If no bundles
      // are specified, compute a graph of bundles starting with previously updated
      // or uninstalled ones. Any bundle that imports a package that is currently
      // exported by a bundle in the graph is added to the graph. The graph is fully
      // contructed when there is no bundles outside the graph that imports a package
      // from a bundle in the graph. The graph may contain UNINSTALLED bundles that are
      // currently still exporting packages.
      BundleImpl[] graph = buildGraph( bundles );
      // 2. Each bundle in the graph will be stopped as described in the Bundle.stop
      // method.
      //
      // 3. Each bundle in the graph that is in the RESOLVED state is moved to the
      // INSTALLED state. The effect of this step is that bundles in the graph are no
      // long RESOLVED.
      Vector activeBundles = new Vector();
      for ( int i = 0; i < graph.length; i++ ) {
         // determine if we should attempt to restart the bundle at the end of the
         // refresh operation
         if ( graph[ i ].getState() == graph[ i ].ACTIVE || graph[ i ].getState() == graph[ i ].STARTING ) {
            activeBundles.addElement( graph[ i ] );
         }

         // ok to call stop on a bundle in any state, except uninstalled (which will
         // cause an exception)
         //
         // special treatment for system bundle which should not be stopped
         if ( graph[ i ].getState() != graph[ i ].UNINSTALLED ) {
            try {
               graph[ i ].stop( true );
            }
            catch ( Throwable e ) {
               // For any exceptions that are thrown, a FrameworkEvent
               // of type ERROR is broadcast, containing the exception.
               cube.fireFrameworkEvent( FrameworkEvent.ERROR, cube.getBundle( 0 ), e );
            }
         }

         // now all bundles in the graph should be uninstalled, installed, or resolved
         //
         // uninstalled bundles need to be unresolved so their exported packages are wiped
         // installed bundle need not be resolved (they have no exported packages to remove
         // and their pending imports are registered)
         // resolved bundle need to be unresolved so their exported packages are wiped
         //
         // could be stopping... what do we do?
         unresolveBundle( (BundleImpl)graph[ i ] );
      }

      // 4. Each bundle in the graph that is in the UNINSTALLED state is removed from
      // the graph and is now completely removed from the framework.
      cube.wipeUninstalledBundles( graph );

      resolveBundles();

      // 5. Each bundle in the graph that was in the ACTIVE state prior to step 2 is
      // started as described in the Bundle.start method, causing all bundles required
      // for the restart to be resolved. It is possible that, as a result of the previous
      // steps, packages that were previously exported no longer are.
      // Threfore, some bundles may be unresolvable until another bundle offering a
      // compatible package for export has been installed in the Framework.
      //
      Bundle bundle;
      for ( int i = 0; i < activeBundles.size(); i++ ) {
         bundle = (Bundle)activeBundles.elementAt( i );
         if ( bundle.getState() == bundle.RESOLVED ) {
            try {
               bundle.start();
            }
            catch ( Throwable e ) {
               // For any exceptions that are thrown, a FrameworkEvent
               // of type ERROR is broadcast, containing the exception.
               cube.fireFrameworkEvent( FrameworkEvent.ERROR, cube.getBundle( 0 ), e );
            }
         }
      }
   }

   // compute a graph of bundles starting with the specified ones. If no bundles
   // are specified, compute a graph of bundles starting with previously updated
   // or uninstalled ones. Any bundle that imports a package that is currently
   // exported by a bundle in the graph is added to the graph. The graph is fully
   // contructed when there is no bundles outside the graph that imports a package
   // from a bundle in the graph. The graph may contain UNINSTALLED bundles that are
   // currently still exporting packages.
   private BundleImpl[] buildGraph( Bundle[] bundles ) {
      // Start the graph
      Vector graph = new Vector();
      if ( bundles != null ) {
         // if specified bundles not null, start graph with specified bundles
         for ( int i = 0; i < bundles.length; i++ ) {
            graph.addElement( bundles[ i ] );
         }
      }
      else {
         // if specified bundles null, start graph with bundles that were previously
         // updated or uninstalled
         // any bundle that exports a package and that has been updated or uninstalled
         // can be found by traversing the list of exporetd packages
         // and seeing which ones have their removePending flag set
         // (we take care of bundles that do not export packages later)
         Enumeration enum = exportedPackages.elements();
         ExportedPackageImpl exportedPackage;
         while ( enum.hasMoreElements() ) {
            exportedPackage = (ExportedPackageImpl)enum.nextElement();
            if ( exportedPackage.isRemovalPending() && graph.indexOf( exportedPackage.getExportingBundle() ) == -1 ) {
               graph.addElement( exportedPackage.getExportingBundle() );
            }
         }
         // we must also add to the graph any bundles that are uninstalled and that
         // do not export packages. This is just so that these bundles can be included
         // in the list of bundles to completely wipe from the framework
         Bundle[] all = cube.getBundles();
         for ( int i = 0; i < all.length; i++ ) {
            if ( all[ i ].getState() == all[ i ].UNINSTALLED && graph.indexOf( all[ i ] ) == -1 ) {
               graph.addElement( all[ i ] );
            }
         }
         // we don't really care here about bundles that have been updated and that
         // do not export packages. If do not import packages exported by bundles in the
         // graph we don't need to do anything with those bundles. If on the other hand
         // they do import packages exported by bundles in the graph, then they will
         // be included in the graph later
      }

      // Build the graph
      boolean done = false;
      while ( !done ) {
         // for each package exported by a bundle in the graph, see if there is an
         // importing bundle that is not in the graph
         Vector bundlesToAdd = new Vector();
         for ( int i = 0; i < graph.size(); i++ ) {
            Vector exports = getProposedExports( (Bundle)graph.elementAt( i ) );
            for ( int j = 0; j < exports.size(); j++ ) {
               PackageExport export = (PackageExport)exports.elementAt( j );
               Enumeration enum = requiredImports.keys();
               while ( enum.hasMoreElements() ) {
                  Bundle bundle = (Bundle)enum.nextElement();
                  Vector imports = getRequiredImports( bundle );
                  for ( int k = 0; k < imports.size(); k++ ) {
                     PackageImport packageImport = (PackageImport)imports.elementAt( k );
                     if ( packageImport.getName().equals( export.getName() ) && graph.indexOf( bundle ) == -1 && bundlesToAdd.indexOf( bundle ) == -1 ) {
                        bundlesToAdd.addElement( bundle );
                     }
                  }
               }
            }
         }

         if ( bundlesToAdd.size() > 0 ) {
            for ( int i = 0; i < bundlesToAdd.size(); i++ ) {
               graph.addElement( bundlesToAdd.elementAt( i ) );
            }
         }
         else {
            done = true;
         }
      }

      BundleImpl[] ret = new BundleImpl[ graph.size() ];
      graph.copyInto( ret );
      return ret;
   }

   private synchronized void unresolveBundle( BundleImpl bundle ) {
      // if the bundle is no longer resolved, it must not import packages
      // also, if the bundle is no longer resolved, it must not export packages
      Enumeration enum = exportedPackages.elements();
      while ( enum.hasMoreElements() ) {
         ExportedPackageImpl exportedPackage = (ExportedPackageImpl)enum.nextElement();
         if ( exportedPackage.getExportingBundle().getBundleId() == bundle.getBundleId() ) {
            // if the package is exported by the bundle itself, remove it from the
            // list of exported bundles (unresolved bundles do not participate in
            // package sharing), and set its stale flag
            exportedPackage.setStale();
            exportedPackages.remove( exportedPackage.getName() );
         }
         else {
            // if the package is exported by another bundle, and the bundle being
            // unresolve imports it, remove it from the list of importers of the package
            // (unresolved bundles do not participate in package sharing)
            //
            // this call will have no effect if bundle is not an importer of
            // exportedPackage (no need to check explicitly if bundle is an importer of
            // the package)
            exportedPackage.removeImportingBundle( bundle );
         }
      }

      if ( bundle.getState() == bundle.RESOLVED ) {
         bundle.setInstalled();
      }
   }

// *****************************************************************************
//
// private utility methods
//
// *****************************************************************************

   private Vector getProposedExports( Bundle bundle ) {
      if ( proposedExports.get( bundle ) == null ) {
         proposedExports.put( bundle, new Vector() );
      }
      return (Vector)proposedExports.get( bundle );
   }

   private Vector getRequiredImports( Bundle bundle ) {
      if ( requiredImports.get( bundle ) == null ) {
         requiredImports.put( bundle, new Vector() );
      }
      return (Vector)requiredImports.get( bundle );
   }

   // Called by cube in the final stage of bundle installation or update, or during
   // package refresh
   // This is the time the list of installed bundles should be traversed to see if any
   // can be resolved as a result of the new bundle installation or update.
   //
   // Resolving bundles is not a simple matter.
   //
   // We can have more than one pending export for a given package name.
   // If the bundle that exports the package with highest version can be resolved,
   // that should be the package that is used. But if that bundle cannot be resolved,
   // we should try to see if any of the bundles exporting lower versions of the
   // package can be resolved, in which case the lower package version will be used.
   //
   // To make things worse, we can have bundles that have cross package dependencies,
   // meaning that some groups of bundles are either all resolved, or none of them is.
   //
   public synchronized void resolveBundles() {
      // Determine the list of bundles that are installed (i.e. waiting to be resolved)
      Bundle[] bundles = cube.getBundles();
      Vector potentiallyResolvedBundles = new Vector();
      for ( int i = 0; i < bundles.length; i++ ) {
         if ( bundles[ i ].getState() == bundles[ i ].INSTALLED ) {
            potentiallyResolvedBundles.addElement( bundles[ i ] );
         }
      }

      boolean done = false;
      Vector notResolvable;
      while ( !done ) {
         notResolvable = simulatePackageExportsFor( potentiallyResolvedBundles );
         if ( notResolvable.size() == 0 ) {
            done = true;
         }
         else {
            for ( int i = 0; i < notResolvable.size(); i++ ) {
               potentiallyResolvedBundles.removeElement( notResolvable.elementAt( i ) );
            }
         }
      }

      // At each iteration of the above loop, either notResolvable.size() == 0, in which
      // case the loop is exited, or notResolvable.size() > 0, in which case we
      // start again with a smaller potentiallyResolvedBundles vector.
      // If potentiallyResolvedBundles.size() reaches zero, then the vector that will
      // be returned from simulatePackageExportsFor() will have zero size and the
      // loop will exit.
      // Therefore, the loop must end in finite time.
      //
      // When the loop ends, the bundles in the potentiallyResolvedBundles vector are
      // the bundles that can be resolved at this stage.

      // Resolve the bundles that can be resolved at this stage
      //
      // First, handle exported packages
      //    create real package exports for all proposed exports (except the proposed
      //    exports with the same package name as a package that is already exported)
      Hashtable toExport = new Hashtable();
      for ( int i = 0; i < potentiallyResolvedBundles.size(); i++ ) {
         Bundle bundle = (Bundle)potentiallyResolvedBundles.elementAt( i );
         Vector proposedExports = getProposedExports( bundle );
         for ( int j = 0; j < proposedExports.size(); j++ ) {
            PackageExport proposedExport = (PackageExport)proposedExports.elementAt( j );
            ExportedPackageImpl alreadyExported = (ExportedPackageImpl)exportedPackages.get( proposedExport.getName() );
            boolean export = false;
            if ( alreadyExported != null ) {
               if ( !alreadyExported.isOpen() ) {
                  // if a package with the same name is already exported but not yet open, we still
                  // have a chance to upgrade it before it gets opened
                  GoRoCoVersion proposedVersion = new GoRoCoVersion( proposedExport.getSpecificationVersion() );
                  GoRoCoVersion alreadyExportedVersion = new GoRoCoVersion( alreadyExported.getSpecificationVersion() );
                  if ( alreadyExportedVersion.isUndefined() && !proposedVersion.isUndefined() ) {
                     export = true;
                  }
                  else {
                     // if alreadyExportedVersion is not undefined and proposedVersion
                     // is undefined, do not export proposed version
                     // then alreadyExportedVersion is considered greater
                     // see implementation of GoRoCoVersion.compareTo()
                     if ( alreadyExportedVersion.compareTo( proposedVersion ) < 0 ) {
                        export = true;
                     }
                  }
               }
            }
            else {
               PackageExport alreadyInList = (PackageExport)toExport.get( proposedExport.getName() );
               if ( alreadyInList == null ) {
                  export = true;
               }
               else {
                  // if several proposed exports found with the same name, take the one
                  // with highest version
                  GoRoCoVersion version1 = new GoRoCoVersion( alreadyInList.getSpecificationVersion() );
                  GoRoCoVersion version2 = new GoRoCoVersion( proposedExport.getSpecificationVersion() );
                  // if either version1 or version2 (but not both) is undefined
                  // version1 is considered greater
                  // see implementation of GoRoCoVersion.compareTo()
                  if ( version1.compareTo( version2 ) < 0 ) {
                     export = true;
                  }
               }
            }
            if ( export ) {
               toExport.put( proposedExport.getName(), proposedExport );
            }
         }
      }
      Enumeration enum = toExport.elements();
      while ( enum.hasMoreElements() ) {
         // create real package export
         PackageExport packageExport = (PackageExport)enum.nextElement();
         ExportedPackageImpl exportedPackage = new ExportedPackageImpl( packageExport.getName(), packageExport.getSpecificationVersion(), packageExport.getExportingBundle(), packageExport.getClassLoader() );
         ExportedPackageImpl unopenedExistingExportBeingUpgraded = (ExportedPackageImpl)exportedPackages.get( packageExport.getName() );
         if ( unopenedExistingExportBeingUpgraded != null ) {
            Bundle[] importingBundles = unopenedExistingExportBeingUpgraded.getImportingBundles();
            for ( int i = 0; i < importingBundles.length; i++ ) {
               exportedPackage.addImportingBundle( importingBundles[ i ] );
            }
            unopenedExistingExportBeingUpgraded.setStale();
         }
         exportedPackages.put( packageExport.getName(), exportedPackage );
      }

      // Second, handle imported packages
      //    transform required imports into real package imports
      //
      // Finally, set the bundle's state to resolved
      for ( int i = 0; i < potentiallyResolvedBundles.size(); i++ ) {
         BundleImpl bundle = (BundleImpl)potentiallyResolvedBundles.elementAt( i );
         Vector requiredImports = getRequiredImports( bundle );
         for ( int j = 0; j < requiredImports.size(); j++ ) {
            PackageImport requiredImport = (PackageImport)requiredImports.elementAt( j );
            // add bundle to the list of importers of the exported package
            ( (ExportedPackageImpl)exportedPackages.get( requiredImport.getName() ) ).addImportingBundle( bundle );
         }
         bundle.setResolved();
      }
   }

   // Pretend that all the specified bundles can be resolved.
   // Then, we get a simulated list of exported packages made up of all the already
   // exported packages and the proposed exports of those bundles (except for the proposed
   // exports that have a package name of a package that is already exported).
   // See, in that situation, which bundles (if any) still miss packages in order to
   // be resolved, and return the list of those packages.
   //
   // If the returned list is empty, then all the specified bundles can be resolved at
   // this stage.
   // If the returned list is not empty, then the bundles in the returned list definitely
   // cannot be resolved at this stage, and there is a question mark over the other
   // bundles in the specified list.
   private Vector simulatePackageExportsFor( Vector bundles ) {
      // Build simulated list of exported packages
      Hashtable potentialExports = new Hashtable();
      for ( int i = 0; i < bundles.size(); i++ ) {
         Bundle bundle = (Bundle)bundles.elementAt( i );
         Vector proposedExports = getProposedExports( bundle );
         for ( int j = 0; j < proposedExports.size(); j++ ) {
            PackageExport proposedExport = (PackageExport)proposedExports.elementAt( j );
            ExportedPackageImpl alreadyExported = (ExportedPackageImpl)exportedPackages.get( proposedExport.getName() );
            boolean addToPotentialExports = false;
            if ( alreadyExported != null ) {
               if ( !alreadyExported.isOpen() ) {
                  // if a package with the same name is already exported but not yet open, we still
                  // have a chance to upgrade it before it gets opened
                  GoRoCoVersion proposedVersion = new GoRoCoVersion( proposedExport.getSpecificationVersion() );
                  GoRoCoVersion alreadyExportedVersion = new GoRoCoVersion( alreadyExported.getSpecificationVersion() );
                  if ( alreadyExportedVersion.isUndefined() && !proposedVersion.isUndefined() ) {
                     addToPotentialExports = true;
                  }
                  else {
                     // if alreadyExportedVersion is not undefined and proposedVersion
                     // is undefined, do not export proposed version
                     // then alreadyExportedVersion is considered greater
                     // see implementation of GoRoCoVersion.compareTo()
                     if ( alreadyExportedVersion.compareTo( proposedVersion ) < 0 ) {
                        addToPotentialExports = true;
                     }
                  }
               }
            }
            else {
               PackageExport alreadyInList = (PackageExport)potentialExports.get( proposedExport.getName() );
               if ( alreadyInList == null ) {
                  addToPotentialExports = true;
               }
               else {
                  // if we have several proposed exports with the same package name, use the
                  // package with the highest specification version
                  GoRoCoVersion version1 = new GoRoCoVersion( alreadyInList.getSpecificationVersion() );
                  GoRoCoVersion version2 = new GoRoCoVersion( proposedExport.getSpecificationVersion() );
                  // if either version1 or version2 (but not both) is undefined
                  // version1 is considered greater
                  // see implementation of GoRoCoVersion.compareTo()
                  if ( version1.compareTo( version2 ) < 0 ) {
                     addToPotentialExports = true;
                  }
               }
            }
            if ( addToPotentialExports ) {
               potentialExports.put( proposedExport.getName(), proposedExport );
            }
         }
      }
      // The simulated list of exported packages contains all the packages that are
      // currently exported, as well as the potential exports
      // The potential exports contain all the package exports proposed by the
      // potentially resolved bundles (but only the highest version of the package
      // if there are several proposed exports for a given package), except for exports
      // of packages with the same name as a package that is already exported
      //
      // There is one exception which makes it possible for the potential exports to
      // contain a package export for a package that is already exported: if the package
      // that is already exported is not open (i.e. noone has loaded classes from that
      // bundle) and if the proposed export has a greater version that the currently
      // exported package.
      Hashtable simulatedExportedPackages = (Hashtable)exportedPackages.clone();
      Enumeration enum = potentialExports.elements();
      while ( enum.hasMoreElements() ) {
         PackageExport potentialExport = (PackageExport)enum.nextElement();
         simulatedExportedPackages.put( potentialExport.getName(), potentialExport );
      }

      // Having obtained the simulated list of exported packages, see which bundles
      // still cannot be resolved
      Vector bundlesStillNotResolved = new Vector();
      for ( int i = 0; i < bundles.size(); i++ ) {
         Bundle bundle = (Bundle)bundles.elementAt( i );
         Vector requiredImports = getRequiredImports( bundle );
         for ( int j = 0; j < requiredImports.size(); j++ ) {
            PackageImport required = (PackageImport)requiredImports.elementAt( j );
            HasVersion available = (HasVersion)simulatedExportedPackages.get( required.getName() );
            if ( available == null ) {
//System.out.println( bundle.getBundleId() + " not resolved - missing: " + required.getName() );
               bundlesStillNotResolved.addElement( bundle );
               break;
            }
            else {
               GoRoCoVersion requiredVersion = new GoRoCoVersion( required.getSpecificationVersion() );
               GoRoCoVersion availableVersion = new GoRoCoVersion( available.getSpecificationVersion() );
               // if either available version or required version (but not both) is undefined
               // available version is considered greater
               // see implementation of GoRoCoVersion.compareTo()
               if ( availableVersion.compareTo( requiredVersion ) < 0 ) {
//System.out.println( bundle.getBundleId() + " not resolved - " + required.getName() + " v" + reqyuredVersion.toString() + " required but v" + availableVersion.toString() + " available" );
                  bundlesStillNotResolved.addElement( bundle );
                  break;
               }
            }
         }
      }

      // The bundles in bundlesStillNotResolved can definitely not be resolved at this
      // stage
      return bundlesStillNotResolved;
   }

}
