/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube.serviceRegistry;

import com.opensugar.cube.AbstractCube;
import com.opensugar.cube.BundleImpl;
import com.opensugar.cube.ServiceRegistrationImpl;
import com.opensugar.cube.ServiceReferenceImpl;
import com.opensugar.cube.UnreachableCodeException;
import com.opensugar.cube.ldap.LDAPFilter;

import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.service.packageadmin.PackageAdmin;
import org.osgi.service.permissionadmin.PermissionAdmin;

import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

public class ServiceRegistry {

   private AbstractCube cube;

   // list of services in the registry    provider bundle --> vector of service registry entries
   private Hashtable bundle_services;

   public ServiceRegistry( AbstractCube cube ) {
      this.cube = cube;
      bundle_services = new Hashtable();
   }

   // Register a service in the registry.
   public ServiceRegistrationImpl registerService( BundleImpl serviceProvider, String[] classNames, Object service, Dictionary properties ) {
      cube.log( cube.LOG_DEBUG, serviceProvider.getLocationWithoutAdminPermissionCheck() + " registering service: " + classNames[ 0 ] );

      // no bundle other than system bundle may register PackageAdmin or PermissionAdmin services
      if ( serviceProvider.getBundleId() != 0 ) {
         for ( int i = 0; i < classNames.length; i++ ) {
            if ( classNames[ i ].equals( PackageAdmin.class.getName() ) ) {
               throw new IllegalArgumentException( "Only system bundle may register PackageAdmin service" );
            }
            else if ( classNames[ i ].equals( PermissionAdmin.class.getName() ) ) {
               throw new IllegalArgumentException( "Only system bundle may register PermissionAdmin service" );
            }
         }
      }

      ServiceRegistrationImpl serviceRegistration = new ServiceRegistrationImpl( serviceProvider, classNames, properties );
      ServiceReferenceImpl serviceReference = (ServiceReferenceImpl)serviceRegistration.getReference();
      ServiceRegistryEntry serviceRegistryEntry = new ServiceRegistryEntry( serviceRegistration, serviceReference, service );
      getServiceRegistryEntries( serviceProvider, true ).addElement( serviceRegistryEntry );
      return serviceRegistration;
   }

   // Return the services that implement the specified class and that pass the specified filter.
   public ServiceReferenceImpl[] getServiceReferences( String className, LDAPFilter filter ) {
      Vector serviceReferences = new Vector();
      Enumeration enum = bundle_services.elements();
      Vector serviceRegistryEntries;
      ServiceRegistryEntry serviceRegistryEntry;
      while ( enum.hasMoreElements() ) {
         serviceRegistryEntries = (Vector)enum.nextElement();
         for ( int i = 0; i < serviceRegistryEntries.size(); i++ ) {
            serviceRegistryEntry = (ServiceRegistryEntry)serviceRegistryEntries.elementAt( i );
            // Return only services that are still registered and that pass the filter (if a filter is specified)
            if ( serviceRegistryEntry.isServiceStillRegistered() && doFilter( serviceRegistryEntry, filter ) && serviceRegistryEntry.isImplemented( className ) ) {
               serviceReferences.addElement( serviceRegistryEntry.getReference() );
            }
         }
      }

      if ( serviceReferences.size() == 0 ) {
         return null;
      }

      ServiceReferenceImpl[] ret = new ServiceReferenceImpl[ serviceReferences.size() ];
      serviceReferences.copyInto( ret );
      return ret;
   }

   // Return the service object for the specified user bundle.
   // This method throws an IllegalStateException if the service of the ServiceReference has been unregistered.
   // This method throws an ServiceRegistryException if the service created by the service factory (when a
   // service factory is used) is not an instance of all the classes named when the service was registered,
   // if the service factory throws an exception while creating the service instance, or if any other problem
   // occurs while creating a new service instance.
   public Object getService( ServiceReference reference, BundleImpl userBundle ) throws ServiceRegistryException {
      Object serviceClass = reference.getProperty( "objectClass" );
      if ( serviceClass instanceof Object[] ) {
         serviceClass = ( (Object[])serviceClass )[ 0 ];
      }
      cube.log( cube.LOG_DEBUG, userBundle.getLocationWithoutAdminPermissionCheck() + " getting service: " + serviceClass );

      Enumeration enum = bundle_services.elements();
      Vector serviceRegistryEntries;
      ServiceRegistryEntry serviceRegistryEntry;
      while ( enum.hasMoreElements() ) {
         serviceRegistryEntries = (Vector)enum.nextElement();
         for ( int i = 0; i < serviceRegistryEntries.size(); i++ ) {
            serviceRegistryEntry = (ServiceRegistryEntry)serviceRegistryEntries.elementAt( i );
            // throw an IllegalStateException if service is unregistered
            if ( serviceRegistryEntry.getReference().equals( reference ) ) {
               // serviceRegistry.getService() throws an IllegalStateException if the service has been
               // unregistered
               // serviceRegistryEntry.getService() throws a ServiceRegistryException if a problem occurs
               // while creating the service instance using the service factory
               return serviceRegistryEntry.getService( userBundle );
            }
         }
      }
      return null;
   }

   // Register that the specified bundle is releasing the service.
   public boolean ungetService( ServiceReference reference, BundleImpl userBundle ) {
      Object serviceClass = reference.getProperty( "objectClass" );
      if ( serviceClass instanceof Object[] ) {
         serviceClass = ( (Object[])serviceClass )[ 0 ];
      }
      cube.log( cube.LOG_DEBUG, userBundle.getLocationWithoutAdminPermissionCheck() + " ungetting service: " + serviceClass );

      Enumeration enum = bundle_services.elements();
      Vector serviceRegistryEntries;
      ServiceRegistryEntry serviceRegistryEntry;
      while ( enum.hasMoreElements() ) {
         serviceRegistryEntries = (Vector)enum.nextElement();
         for ( int i = 0; i < serviceRegistryEntries.size(); i++ ) {
            serviceRegistryEntry = (ServiceRegistryEntry)serviceRegistryEntries.elementAt( i );
            if ( serviceRegistryEntry.getReference().equals( reference ) ) {
               return serviceRegistryEntry.ungetService( userBundle );
            }
         }
      }

      // This should never happen
      // ServiceReference instances are all created by this class and immediately stored in
      // a service registry entry.  So any reference passed to this method should be found when
      // looking through all service registry entries
      throw new UnreachableCodeException();
   }

   // Return the list of services registered by the specified bundle.
   // Only include services that are still registered.
   public ServiceRegistrationImpl[] getServicesRegisteredBy( BundleImpl bundle ) {
      Vector serviceRegistrations = new Vector();
      Enumeration enum = bundle_services.elements();
      Vector serviceRegistryEntries;
      ServiceRegistryEntry serviceRegistryEntry;
      while ( enum.hasMoreElements() ) {
         serviceRegistryEntries = (Vector)enum.nextElement();
         for ( int i = 0; i < serviceRegistryEntries.size(); i++ ) {
            serviceRegistryEntry = (ServiceRegistryEntry)serviceRegistryEntries.elementAt( i );
            // Only return services that are still registered.
            if ( serviceRegistryEntry.isServiceStillRegistered() && serviceRegistryEntry.isProvidedBy( bundle ) ) {
               serviceRegistrations.addElement( serviceRegistryEntry.getRegistration() );
            }
         }
      }

      if ( serviceRegistrations.size() == 0 ) {
         return null;
      }

      ServiceRegistrationImpl[] ret = new ServiceRegistrationImpl[ serviceRegistrations.size() ];
      serviceRegistrations.copyInto( ret );
      return ret;
   }

   // Return list of services used by the specified bundle.
   // Include used services even if they are unregistered.
   public ServiceReferenceImpl[] getServicesUsedBy( BundleImpl bundle ) {
      Vector serviceReferences = new Vector();
      Enumeration enum = bundle_services.elements();
      Vector serviceRegistryEntries;
      ServiceRegistryEntry serviceRegistryEntry;
      while ( enum.hasMoreElements() ) {
         serviceRegistryEntries = (Vector)enum.nextElement();
         for ( int i = 0; i < serviceRegistryEntries.size(); i++ ) {
            serviceRegistryEntry = (ServiceRegistryEntry)serviceRegistryEntries.elementAt( i );
            // Return service even if they have been unregistered
            if ( serviceRegistryEntry.isUsedBy( bundle ) ) {
               serviceReferences.addElement( serviceRegistryEntry.getReference() );
            }
         }
      }

      if ( serviceReferences.size() == 0 ) {
         return null;
      }

      ServiceReferenceImpl[] ret = new ServiceReferenceImpl[ serviceReferences.size() ];
      serviceReferences.copyInto( ret );
      return ret;
   }

   public ServiceRegistryEntry getServiceRegistryEntry( ServiceRegistration serviceRegistration ) {
      Vector serviceReferences = new Vector();
      Enumeration enum = bundle_services.elements();
      Vector serviceRegistryEntries;
      ServiceRegistryEntry serviceRegistryEntry;
      while ( enum.hasMoreElements() ) {
         serviceRegistryEntries = (Vector)enum.nextElement();
         for ( int i = 0; i < serviceRegistryEntries.size(); i++ ) {
            serviceRegistryEntry = (ServiceRegistryEntry)serviceRegistryEntries.elementAt( i );
            if ( serviceRegistryEntry.getRegistration().equals( serviceRegistration ) ) {
               return serviceRegistryEntry;
            }
         }
      }
      // This should never happen, or it means we asked for a service registry entry that does
      // not exist. (ServiceRegistration is the only object that calls this method, and it should
      // never call this method with a service registration that does not correspond to an
      // existing service registry entry.
      throw new UnreachableCodeException();
   }


// *****************************************************************************

   // Return the list of service registry entries provided by the specified bundle.
   // If no list exists for the specified bundle and createIfNull is true, then create a new
   // empty list of services for the bundle.
   private Vector getServiceRegistryEntries( BundleImpl serviceProvider, boolean createIfNull ) {
      if ( bundle_services.get( serviceProvider ) == null && createIfNull ) {
         bundle_services.put( serviceProvider, new Vector() );
      }
      return (Vector)bundle_services.get( serviceProvider );
   }

   // Determine if the specified service registry entry passses the filter or not.
   // Return true if the filter is null.
   private boolean doFilter( ServiceRegistryEntry serviceRegistryEntry, LDAPFilter filter ) {
      if ( filter != null ) {
         try {
            LDAPFilter caseInsensitiveFilter = new LDAPFilter( filter.getFilterString(), false );
            return caseInsensitiveFilter.match( serviceRegistryEntry.getReference() );
         }
         catch ( InvalidSyntaxException e ) {
            // creating a filter out of the filter string of another filter
            // so filter string syntax must already have been checked
            throw new UnreachableCodeException();
         }
      }
      return true;
   }

}