/*
 * JEFFREE: Java(TM) Embedded Framework FREE
 * Copyright (C) 1999-2003 - Opensugar
 *
 * The contents of this file are subject to the Jeffree Public License,
 * as defined by the file JEFFREE_LICENSE.TXT 
 * 
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the Objectweb web site
 * (www.objectweb.org). 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License. 
 * 
 * The Original Code is JEFFREE, including the java package com.opensugar.cube,
 * released January 1, 2003.
 * 
 * The Initial Developer of the Original Code is Opensugar.
 * The Original Code is Copyright Opensugar.
 * All Rights Reserved.
 *
 * Initial developer(s): Pierre Scokaert (Opensugar)
 * Contributor(s):
 */

package com.opensugar.cube.serviceRegistry;

import com.opensugar.cube.BundleImpl;
import com.opensugar.cube.ServiceReferenceImpl;

import org.osgi.framework.Bundle;
import org.osgi.framework.ServiceRegistration;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceFactory;
import org.osgi.framework.Constants;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Hashtable;

// An entry of the service registry.
// Handles users of the service.
public class ServiceRegistryEntry {

   // the service registration for this service
   private ServiceRegistration registration;
   // the service object for this service (may be the service itself or a factory for the service)
   private Object service;
   // the registered users of this service
   private Vector users;
   // the service reference for this service
   // Note: we need to keep a local copy service reference so that we can still access it after
   // the service is unregistered (after service is unregistered, calling getReference on service
   // registration throws an IllegalStateException)
   private ServiceReferenceImpl reference;

   // a cache of service instances     // user bundle --> cached instance of service
   private Hashtable cachedServiceInstances;

   public ServiceRegistryEntry( ServiceRegistration registration, ServiceReferenceImpl reference, Object service ) {
      this.registration = registration;
      this.reference = reference;
      this.service = service;

      users = new Vector();
      cachedServiceInstances = new Hashtable();
   }

   // Return the service registration for this service.
   protected ServiceRegistration getRegistration() {
      return registration;
   }

   // Return the service reference for this service.
   protected ServiceReferenceImpl getReference() {
      return reference;
   }

   // Return a service object for the specified user bundle.
   // This method throws an IllegalStateException if the service has been unregistered.
   // This method throws a ServiceRegistryException if the service factory throws an
   // exception while creating a new service intance, or if the service created by the service
   // factory is not an instance of all the classes names during this service's registration.
   protected synchronized Object getService( Bundle userBundle ) throws ServiceRegistryException {
      // call registration's getReference method so that an illegal state exception is thrown if the service has been unregistered
      registration.getReference();

      Object serviceInstance;
      if ( service instanceof ServiceFactory ) {
         // service is a service factory
         if ( cachedServiceInstances.get( userBundle ) != null ) {
            // a service instance is already cached for this bundle
            return cachedServiceInstances.get( userBundle );
         }
         else {
            // no service instance is cached for this bundle.
            // create one
            try {
               serviceInstance = ( (ServiceFactory)service ).getService( userBundle, registration );
            }
            catch( Exception e ) {
               throw new ServiceRegistryException( "Exception occured while calling service factory", e );
            }

            if ( serviceInstance == null ) {
               throw new ServiceRegistryException( "Service factory returned a null object" );
            }

            // check that service instance is an instance of all the classes named during this service's registration
            // if that is not the case, throw an exception
            try {
               String[] objectClass = (String[])reference.getProperty( "objectClass" );
               BundleImpl bundle;
               for ( int i = 0; i < objectClass.length; i++ ) {
                  bundle = (BundleImpl)reference.getBundle();
                  if ( !bundle.getClassLoader().loadClass( objectClass[ i ] ).isInstance( serviceInstance ) ) {
                     throw new ServiceRegistryException( "Invalid service type created by service factory" );
                  }
               }
            }
            catch( ClassNotFoundException e ) {
               throw new ServiceRegistryException( "Class not found while checking service created by service factory", e );
            }

            // cache new service instance
            cachedServiceInstances.put( userBundle, serviceInstance );
         }
      }
      else {
         // service is a service instance (not a factory)
         serviceInstance = service;
      }
      // record new user bundle (thus incrementing bundle use count)
      addUser( userBundle );
      // return service instance
      return serviceInstance;
   }

   // Release the service used by the specified bundle.
   protected synchronized boolean ungetService( Bundle userBundle ) {
      return removeUser( userBundle );
   }

   // Determine if the specified class is one of the classes this service is registered under.
   protected boolean isImplemented( String className ) {
      if ( className == null ) {
         return true;
      }

      String[] classNames = (String[])reference.getProperty( Constants.OBJECTCLASS );
      for ( int i = 0; i < classNames.length; i++ ) {
         if ( classNames[ i ].equals( className ) ) {
            return true;
         }
      }
      return false;
   }

   // Determine if this service is provided by the specified bundle.
   protected boolean isProvidedBy( Bundle bundle ) {
      if ( reference.getBundle().getBundleId() == bundle.getBundleId() ) {
         return true;
      }
      return false;
   }

   // Determine if this service is currently used by the specified bundle.
   protected synchronized boolean isUsedBy( Bundle bundle ) {
      if ( users.indexOf( bundle ) != -1 ) {
         return true;
      }
      return false;
   }

   // Determine if this service is still registered.
   protected boolean isServiceStillRegistered() {
      try {
         // If this service is no longer registered, calling getReference() on the
         // service registration throws an IllegalStateException.
         registration.getReference();
         return true;
      }
      catch( IllegalStateException e ) {
         return false;
      }
   }

   // This method is called by ServiceRegistrationImpl when the service is unregistered.
   // When that happens, we need to do the following:
   // For each bundle whose use count for this service is greater than zero,
   //    1. set the bundle's use count for this service to zero
   //    2. if the service was registered with a service factory, call the factory's
   //       unget method to release the service object for the bundle
   public synchronized void serviceUnregistered() {
      while ( users.size() > 0 ) {
         removeUser( (Bundle)users.elementAt( 0 ) );
      }
   }

// *****************************************************************************

   // Register a bundle as a user of this service.
   private synchronized void addUser( Bundle user ) {
      if ( users.indexOf( user ) == -1 ) {
         // If the bundle is not yet a user of the service, let the service reference
         // known the user bundle is now using it.
         // The service reference needs to know because it needs to return the
         // list of user bundles in its getUsingBundles() method
         reference.addUsingBundle( user );
      }

      // Allow a user bundle to be added multiple times to the vector of users.
      // Thus a bundle's use count for the service is obtained by counting the number
      // of times the bundle appears in the list of users.
      users.addElement( user );
   }

   // Unregister a bundle as a user of this service.
   // Return true if the specified bundle was a registered user of this service, false
   // otherwise.
   private synchronized boolean removeUser( Bundle user ) {
      // Remove a single occurence of user from the vector of users.
      // Thus if a bundle has got the service multiple times, it needs to unget it
      // multiple times before it disappears from the list of users.
      if ( users.indexOf( user ) == -1 ) {
         return false;
      }

      users.removeElement( user );
      if ( users.indexOf( user ) == -1 ) {
         // This bundle is no longer a user of this service.
         // Release service object
         if ( service instanceof ServiceFactory ) {
            Object serviceInstance = cachedServiceInstances.get( user );
            ( (ServiceFactory)service ).ungetService( user, registration, serviceInstance );
         }
         // Remove cached service object if one was created.
         cachedServiceInstances.remove( user );

         // If the bundle is no longer using this service, let the service reference
         // known.
         // The service reference needs to know because it needs to return the
         // list of user bundles in its getUsingBundles() method
         reference.removeUsingBundle( user );
      }

      return true;
   }

}